unit OptionsFrm;

interface

uses
  FMX.Forms, FMX.Edit, FMX.StdCtrls, FMX.Controls, FMX.TabControl,
  System.Classes, FMX.Types, FMX.Controls.Presentation,
  fmxRVStyle, fmxRVEdit, fmxRichView, fmxRVTypes, FMX.Layouts;

{$I RV_FMXDefs.inc}

type
  TfrmOptions = class(TForm)
    PageControl1: TTabControl;
    TabSheet1: TTabItem;
    TabSheet2: TTabItem;
    TabSheet3: TTabItem;
    ltButtons: TFlowLayout;
    btnCancel: TButton;
    btnOk: TButton;
    TopToolBar: TToolBar;
    ToolBarLabel: TLabel;
    btnBack: TSpeedButton;
    btnCancel1: TSpeedButton;
    VertScrollBox1: TVertScrollBox;
    cbRVFSaveBackground: TCheckBox;
    cbRVFSaveLayout: TCheckBox;
    cbRVFBinary: TCheckBox;
    cbRVFSavePictures: TCheckBox;
    cbRVFSaveControls: TCheckBox;
    cbRVFSaveTextStyles: TCheckBox;
    cbRVFSaveParaStyles: TCheckBox;
    VertScrollBox2: TVertScrollBox;
    cbRVFLoadBackground: TCheckBox;
    cbRVFLoadLayout: TCheckBox;
    cbRVFIgnoreUnknownPictures: TCheckBox;
    cbRVFIgnoreUnknownControls: TCheckBox;
    cbRVFInvStyles: TCheckBox;
    cbRVFInvImageIndices: TCheckBox;
    cbRVFIgnoreUnknownControlProps: TCheckBox;
    VertScrollBox3: TVertScrollBox;
    cbHTMLOverrideImages: TCheckBox;
    cbHTMLListAsTest: TCheckBox;
    cbHTMLForceNonTextCSS: TCheckBox;
    cbHTMLDefault0Style: TCheckBox;
    cbHTMLSaveImageSizes: TCheckBox;
    cbHTMLInlineImages: TCheckBox;
    cbHTMLUseCheckpointsNames: TCheckBox;
    Layout1: TLayout;
    Label2: TLabel;
    txtHTMLTitle: TEdit;
    procedure btnBackClick(Sender: TObject);
    procedure btnOkClick(Sender: TObject);
    procedure btnCancel1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; var KeyChar: Char;
      Shift: TShiftState);
  private
    { Private declarations }
    procedure GetOptions(RV: TCustomRichView);
    procedure Apply;
  public
    { Public declarations }
    procedure SetOptions(RV: TCustomRichView);
  end;

var
  frmOptions: TfrmOptions;

implementation

{$R *.FMX}
uses
  Unit1;

{ TfrmOptions }

procedure TfrmOptions.Apply;
begin
  GetOptions(Form1.RichViewEdit1);
end;

procedure TfrmOptions.btnBackClick(Sender: TObject);
begin
  Apply;
  Close;
end;

procedure TfrmOptions.btnCancel1Click(Sender: TObject);
begin
  Close;
end;

procedure TfrmOptions.btnOkClick(Sender: TObject);
begin
  Apply;
end;

procedure TfrmOptions.FormCreate(Sender: TObject);
begin
  {$IFDEF RVMOBILE}
  ltButtons.Visible := False;
  {$ELSE}
  TopToolBar.Visible := False;
  {$ENDIF}
end;

procedure TfrmOptions.FormKeyDown(Sender: TObject; var Key: Word;
  var KeyChar: Char; Shift: TShiftState);
begin
  if Key = rvvkHardwareBack then
    btnBackClick(nil);
end;

procedure TfrmOptions.GetOptions(RV: TCustomRichView);
{..........................................................}
  procedure SetRVFOption(Value: Boolean; Option: TRVFOption);
  begin
    if Value then
      RV.RVFOptions := RV.RVFOptions + [Option]
    else
      RV.RVFOptions := RV.RVFOptions - [Option];
  end;
  {..........................................................}
  procedure SetHTMLImageOption(Value: Boolean; Option: TRVHTMLSaveImageOption);
  begin
    if Value then
      RV.HTMLSaveProperties.ImageOptions := RV.HTMLSaveProperties.ImageOptions + [Option]
    else
      RV.HTMLSaveProperties.ImageOptions := RV.HTMLSaveProperties.ImageOptions - [Option];
  end;
  {..........................................................}

begin;
  SetRVFOption(cbRVFSaveBackground.IsChecked, rvfoSaveBack);
  SetRVFOption(cbRVFSaveLayout.IsChecked, rvfoSaveLayout);
  SetRVFOption(cbRVFBinary.IsChecked, rvfoSaveBinary);
  SetRVFOption(cbRVFSavePictures.IsChecked, rvfoSavePicturesBody);
  SetRVFOption(cbRVFSaveControls.IsChecked, rvfoSaveControlsBody);
  SetRVFOption(cbRVFLoadBackground.IsChecked, rvfoLoadBack);
  SetRVFOption(cbRVFLoadLayout.IsChecked, rvfoLoadLayout);
  SetRVFOption(cbRVFIgnoreUnknownPictures.IsChecked, rvfoIgnoreUnknownPicFmt);
  SetRVFOption(cbRVFIgnoreUnknownControls.IsChecked, rvfoIgnoreUnknownCtrls);
  SetRVFOption(cbRVFIgnoreUnknownControlProps.IsChecked, rvfoIgnoreUnknownCtrlProperties);

  SetRVFOption(cbRVFInvStyles.IsChecked, rvfoConvUnknownStylesToZero);
  SetRVFOption(cbRVFInvImageIndices.IsChecked, rvfoConvLargeImageIdxToZero);
  SetRVFOption(cbRVFSaveTextStyles.IsChecked, rvfoSaveTextStyles);
  SetRVFOption(cbRVFSaveParaStyles.IsChecked, rvfoSaveParaStyles);

  SetHTMLImageOption(cbHTMLOverrideImages.IsChecked, rvhtmlsioOverrideImages);
  SetHTMLImageOption(cbHTMLSaveImageSizes.IsChecked, rvhtmlsioImageSizes);
  SetHTMLImageOption(cbHTMLInlineImages.IsChecked, rvhtmlsioInlineImages);
  if cbHTMLListAsTest.IsChecked then
    RV.HTMLSaveProperties.ListMarkerSavingType:= rvhtmlmstAsText
  else
    RV.HTMLSaveProperties.ListMarkerSavingType:= rvhtmlmstNormal;
  if cbHTMLForceNonTextCSS.IsChecked then
    RV.HTMLSaveProperties.CSSOptions := RV.HTMLSaveProperties.CSSOptions + [rvcsssoForceNonTextCSS]
  else
    RV.HTMLSaveProperties.CSSOptions := RV.HTMLSaveProperties.CSSOptions - [rvcsssoForceNonTextCSS];
  RV.HTMLSaveProperties.Default0Style := cbHTMLDefault0Style.IsChecked;
  RV.HTMLSaveProperties.UseCheckpointNames := cbHTMLUseCheckpointsNames.IsChecked;
  RV.DocParameters.Title := txtHTMLTitle.Text;
end;

{------------------------------------------------------------------------------}
procedure TfrmOptions.SetOptions(RV: TCustomRichView);
begin
  cbRVFSaveBackground.IsChecked := rvfoSaveBack in RV.RVFOptions;
  cbRVFSaveLayout.IsChecked := rvfoSaveLayout in RV.RVFOptions;
  cbRVFBinary.IsChecked := rvfoSaveBinary in RV.RVFOptions;
  cbRVFSavePictures.IsChecked := rvfoSavePicturesBody in RV.RVFOptions;
  cbRVFSaveControls.IsChecked := rvfoSaveControlsBody in RV.RVFOptions;
  cbRVFLoadBackground.IsChecked := rvfoLoadBack in RV.RVFOptions;
  cbRVFLoadLayout.IsChecked := rvfoLoadLayout in RV.RVFOptions;
  cbRVFIgnoreUnknownPictures.IsChecked := rvfoIgnoreUnknownPicFmt in RV.RVFOptions;
  cbRVFIgnoreUnknownControls.IsChecked := rvfoIgnoreUnknownCtrls in RV.RVFOptions;
  cbRVFIgnoreUnknownControlProps.IsChecked := rvfoIgnoreUnknownCtrlProperties in RV.RVFOptions;
  cbRVFInvStyles.IsChecked := rvfoConvUnknownStylesToZero in RV.RVFOptions;
  cbRVFInvImageIndices.IsChecked := rvfoConvLargeImageIdxToZero in RV.RVFOptions;
  cbRVFSaveTextStyles.IsChecked := rvfoSaveTextStyles in RV.RVFOptions;
  cbRVFSaveParaStyles.IsChecked := rvfoSaveParaStyles in RV.RVFOptions;

  cbHTMLOverrideImages.IsChecked := rvhtmlsioOverrideImages in RV.HTMLSaveProperties.ImageOptions;
  cbHTMLSaveImageSizes.IsChecked := rvhtmlsioImageSizes in RV.HTMLSaveProperties.ImageOptions;
  cbHTMLInlineImages.IsChecked := rvhtmlsioInlineImages in RV.HTMLSaveProperties.ImageOptions;
  cbHTMLListAsTest.IsChecked := RV.HTMLSaveProperties.ListMarkerSavingType= rvhtmlmstAsText;
  cbHTMLForceNonTextCSS.IsChecked := rvcsssoForceNonTextCSS in RV.HTMLSaveProperties.CSSOptions;
  cbHTMLDefault0Style.IsChecked := RV.HTMLSaveProperties.Default0Style;
  cbHTMLUseCheckpointsNames.IsChecked := RV.HTMLSaveProperties.UseCheckpointNames;
  txtHTMLTitle.Text := RV.DocParameters.Title;
end;

end.
