unit PropFrm;

interface

uses
  System.UITypes, System.SysUtils,
  FMX.Graphics, FMX.Controls.Presentation, FMX.StdCtrls, FMX.ListBox, FMX.Edit,
  FMX.Controls, FMX.TabControl, System.Classes, FMX.Types, FMX.Forms,
  FMX.Layouts, FMX.ImgList,
  fmxRVTypes, fmxRVItem, fmxRVStyle, fmxRVEdit, fmxRVClasses,
  fmxRVGraphicsFM, fmxRVTable;

{$I RV_FMXDefs.inc}

type
  TfrmProp = class(TForm)
    TabControl1: TTabControl;
    tsBullet: TTabItem;
    tsHotSpot: TTabItem;
    tsPicture: TTabItem;
    tsText: TTabItem;
    lblText: TLabel;
    tsComponent: TTabItem;
    lblComponent: TLabel;
    tsBreak: TTabItem;
    ImageControl1: TImageControl;
    TopToolBar: TToolBar;
    ToolBarLabel: TLabel;
    btnBack: TSpeedButton;
    btnCancel1: TSpeedButton;
    VertScrollBox1: TVertScrollBox;
    Layout1: TLayout;
    Label1: TLabel;
    cmbBulletImage: TComboBox;
    tsCommon: TTabItem;
    VertScrollBox2: TVertScrollBox;
    Layout2: TLayout;
    Label3: TLabel;
    txtName: TEdit;
    Layout3: TLayout;
    Label6: TLabel;
    txtTag: TEdit;
    ltButtons: TFlowLayout;
    btnCancel: TButton;
    btnOk: TButton;
    VertScrollBox3: TVertScrollBox;
    Layout4: TLayout;
    Label12: TLabel;
    cmbHotSpotImage: TComboBox;
    VertScrollBox4: TVertScrollBox;
    Layout5: TLayout;
    Label7: TLabel;
    cmbPicVAlign: TComboBox;
    ltChangePicture: TLayout;
    btnChangePicture: TButton;
    VertScrollBox5: TVertScrollBox;
    Layout7: TLayout;
    Label8: TLabel;
    Layout8: TLayout;
    Label13: TLabel;
    txtWidth: TEdit;
    txtHeight: TEdit;
    Layout9: TLayout;
    Label2: TLabel;
    cmbCtrlVAlign: TComboBox;
    VertScrollBox6: TVertScrollBox;
    Layout11: TLayout;
    Label9: TLabel;
    Layout12: TLayout;
    Label14: TLabel;
    txtBreakWidth: TEdit;
    Layout10: TLayout;
    Label4: TLabel;
    cmbBreakColor: TComboBox;
    cmbBreakStyle: TComboBox;
    procedure btnChangePictureClick(Sender: TObject);
    procedure btnOkClick(Sender: TObject);
    procedure btnBackClick(Sender: TObject);
    procedure btnCancel1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; var KeyChar: Char;
      Shift: TShiftState);
  private
    { Private declarations }
    ImageWidth, ImageHeight: TRVStyleLength;
    ctrl: TControl;
    procedure SetToEditor(rve: TCustomRichViewEdit);
  public
    { Public declarations }
    procedure GetFromEditor(rve: TCustomRichViewEdit);
  end;

var
  frmProp: TfrmProp;

implementation

uses Unit1;

{$R *.FMX}

procedure TfrmProp.btnChangePictureClick(Sender: TObject);
begin
  ImageControl1.ShowOpenDialog;
end;

procedure TfrmProp.GetFromEditor(rve: TCustomRichViewEdit);
var
  ItemName:   TRVUnicodeString;
  Tag:        TRVTag;
  Index:      Integer;
  VAlign:     TRVVAlign;
  ImageList:  TCustomImageList;
  gr:         TRVGraphicFM;
  BreakColor: TAlphaColor;
  BreakStyle: TRVBreakStyle;
  BreakWidth: TRVStyleLength;
begin
  tsBullet.Visible := False;
  tsHotSpot.Visible := False;
  tsPicture.Visible := False;
  tsText.Visible := False;
  tsComponent.Visible := False;
  tsBreak.Visible := False;
  txtName.Enabled := True;
  case rve.CurItemStyle of
    rvsBullet:
      begin
        rve.GetCurrentBulletInfo(ItemName, Index, ImageList, Tag,
          ImageWidth, ImageHeight);
        tsBullet.Visible := True;
        tsBullet.IsSelected := True;
        cmbBulletImage.ItemIndex := Index;
        txtName.Text := ItemName;
        txtTag.Text := Tag;
        cmbBulletImage.SetFocus;
      end;
    rvsHotspot:
      begin
        // you can use GetCurrentBulletInfo or GetCurrentHotspotInfo
        // to receive info about hotspot in caret position.
        // in this demo we do not need HotImageIndex, because here
        // HotImageIndex = ImageIndex+2
        // and so we can use GetCurrentBulletInfo
        rve.GetCurrentBulletInfo(ItemName, Index, ImageList, Tag,
          ImageWidth, ImageHeight);
        tsHotSpot.Visible := True;
        tsHotSpot.IsSelected := True;
        cmbHotspotImage.ItemIndex := Index - 3;
        cmbHotspotImage.SetFocus;
        txtName.Text := ItemName;
        txtTag.Text := Tag;
      end;
    rvsPicture, rvsHotPicture:
      begin
        rve.GetCurrentPictureInfo(ItemName, gr, VAlign, Tag);
        tsPicture.Visible := True;
        tsPicture.IsSelected := True;
        if gr is TRVBitmap then
        ImageControl1.Bitmap := TRVBitmap(gr).Bitmap;
        txtName.Text := ItemName;
        txtTag.Text := Tag;
        cmbPicVAlign.ItemIndex := Integer(VAlign);
        cmbPicVAlign.SetFocus;
      end;
    rvsComponent:
      begin
        rve.GetCurrentControlInfo(ItemName, ctrl, VAlign, Tag);
        tsComponent.Visible := True;
        tsComponent.IsSelected := True;
        txtWidth.Text := FloatToStr(ctrl.Width);
        txtHeight.Text := FloatToStr(ctrl.Height);
        txtName.Text := ItemName;
        lblComponent.Text := ctrl.ClassName;
        txtTag.Text := Tag;
        cmbCtrlVAlign.ItemIndex := Integer(VAlign);
        txtWidth.SetFocus;
      end;
    rvsBreak:
      begin
        tsBreak.Visible := True;
        tsBreak.IsSelected := True;
        rve.GetCurrentBreakInfo(BreakWidth, BreakStyle, BreakColor, Tag);
        txtBreakWidth.Text := IntToStr(BreakWidth);
        case BreakColor of
          TAlphaColorRec.Null:
            cmbBreakColor.ItemIndex := 0;
          TAlphaColorRec.Red:
            cmbBreakColor.ItemIndex := 1;
          TAlphaColorRec.Green:
            cmbBreakColor.ItemIndex := 2;
          TAlphaColorRec.Blue:
            cmbBreakColor.ItemIndex := 3;
        end;
        cmbBreakStyle.ItemIndex := ord(BreakStyle);
        txtName.Text := '(not available for breaks)';
        txtName.Enabled := False;
        txtTag.Text := Tag;
        cmbBreakColor.SetFocus;
      end;
    else
      begin
        tsCommon.IsSelected := True;
        txtName.Text := rve.GetCurrentItemText;
        txtTag.Text := rve.GetCurrentTag;
        txtName.SetFocus;
      end;
  end;
end;
{------------------------------------------------------------------------------}
procedure TfrmProp.SetToEditor(rve: TCustomRichViewEdit);
var
  gr:         TRVGraphicFM;
  BreakColor: TAlphaColor;
  BreakStyle: TRVBreakStyle;
  BreakWidth: TRVStyleLength;
begin
  case rve.CurItemStyle of
    rvsBullet:
      begin
        rve.SetCurrentBulletInfo(txtName.Text,
          cmbBulletImage.ItemIndex, nil, txtTag.Text,
          ImageWidth, ImageHeight);
      end;
    rvsHotspot:
      begin
        rve.SetCurrentHotspotInfo(txtName.Text,
          cmbHotspotImage.ItemIndex + 3,
          cmbHotspotImage.ItemIndex + 3 + 2, nil,
          txtTag.Text, ImageWidth, ImageHeight);
      end;
    rvsPicture, rvsHotPicture:
      begin
        { first we need to create a copy of image ...}
        gr := TRVBitmap.CreateFromBitmap(ImageControl1.Bitmap, True);
        rve.SetCurrentPictureInfo(txtName.Text, gr,
          TRVVAlign(cmbPicVAlign.ItemIndex), txtTag.Text);
      end;
    rvsComponent:
      begin
        // we want these setting to be undone as one action,
        // so we use BeginUndoGroup, SetUndoGroupMode(True), settings, SetUndoGroupMode(False)
        rve.BeginUndoGroup(rvutModifyItem);
        // you can use BeginUndoCustomGroup instead of BeginUndoGroup
        // example:
        // rve.BeginUndoCustomGroup('modifying control');
        // In this case undo type will be rvutCustom
        // (look at TForm1.UpdateUndoMenu in this file)
        rve.SetUndoGroupMode(True);
        rve.SetCurrentControlInfo(txtName.Text,
          TRVVAlign(cmbCtrlVAlign.ItemIndex), txtTag.Text);
        rve.ResizeCurrentControl(
          StrToFloatDef(txtWidth.Text, ctrl.Width),
          StrToFloatDef(txtHeight.Text, ctrl.Height));
        rve.SetUndoGroupMode(False);
      end;
    rvsBreak:
      begin
        case cmbBreakColor.ItemIndex of
          1:
            BreakColor := TAlphaColorRec.Red;
          2:
            BreakColor := TAlphaColorRec.Green;
          3:
            BreakColor := TAlphaColorRec.Blue;
          else
            BreakColor := TAlphaColorRec.Null;
        end;
        BreakWidth := StrToIntDef(txtBreakWidth.Text, 1);
        BreakStyle := TRVBreakStyle(cmbBreakStyle.ItemIndex);
        rve.SetCurrentBreakInfo(BreakWidth, BreakStyle, BreakColor,
          txtTag.Text);
      end;
    else
      begin
        rve.BeginUndoGroup(rvutModifyItem);
        rve.SetUndoGroupMode(True);
        rve.SetCurrentItemText(txtName.Text);
        rve.SetCurrentTag(txtTag.Text);
        rve.SetUndoGroupMode(False);
      end;
  end;
end;
{------------------------------------------------------------------------------}
procedure TfrmProp.btnOkClick(Sender: TObject);
begin
  SetToEditor(Form1.RichViewEdit1);
end;

procedure TfrmProp.FormCreate(Sender: TObject);
begin
  {$IFDEF RVMOBILE}
  ltButtons.Visible := False;
  ltChangePicture.Visible := False;
  {$ELSE}
  TopToolBar.Visible := False;
  {$ENDIF}
end;

procedure TfrmProp.FormKeyDown(Sender: TObject; var Key: Word;
  var KeyChar: Char; Shift: TShiftState);
begin
  if Key = rvvkHardwareBack then
    btnBackClick(nil);
end;

procedure TfrmProp.btnBackClick(Sender: TObject);
begin
  SetToEditor(Form1.RichViewEdit1);
  Close;
end;

procedure TfrmProp.btnCancel1Click(Sender: TObject);
begin
  Close;
end;



end.
