{*******************************************************}
{                                                       }
{       RichView                                        }
{       Editor Demo.                                    }
{       RichView components.                            }
{                                                       }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

{
 This demo uses a predefined set of styles.
 This demo shows how to implement:
 - "Edit" menu (Clipboard and Undo);
 - "Search" command;
 - checkpoints;
 - inserting document or image;
 - inserting some controls;
 - working with table;
 - changing background properties and other options.
 It does not implement:
 - commands like "make bold" or "apply font";
 - standard "File" commands (Open, Save, SaveAs.
 (see Editor 2 demo for these features)

 Platform notes: the demo uses 2 duplicate menus:
 - MainMenu1 for desktop OS
 - Popup1 for Android and iOS
}

{$I RV_FMXDefs.inc}

unit Unit1;

interface

uses
  System.SysUtils, System.Types, System.UITypes, System.Classes,
  FMX.Types, FMX.Controls, FMX.Forms, FMX.Dialogs, FMX.Objects, FMX.Graphics,
  FMX.Controls.Presentation, System.ImageList,
  FMX.ImgList, FMX.Printer, FMX.Menus, FMX.ListBox, FMX.StdCtrls, FMX.Layouts,
  FMX.Edit,
  {$IFDEF OSX}MacApi.Helpers, MacApi.Foundation,{$ENDIF}
  {$IFDEF IOS}System.IOUtils,{$ENDIF}
  {$IFDEF ANDROID}System.IOUtils,{$ENDIF}
  FMX.TabControl, FMX.Effects, System.Generics.Collections, FMX.Filter.Effects,
  FMX.Platform, FMX.VirtualKeyboard,
  fmxRVStyle, fmxRVScroll, fmxRichView, fmxRVItem, fmxRVTable, fmxRVTypes,
  fmxCRVData, fmxCRVFData, fmxRVRVData, fmxRVEdit, fmxRVGraphicsFM,
  fmxRVGrInFM, fmxRVUni, fmxRVGrHandler, fmxRVGrIn, fmxRVClipboard,
  fmxRVClipboardFM, fmxRVDialogsFM, fmxRVMessages, fmxRVFileOpenFM,
  fmxRVClasses, fmxRVCanvasFM,
  TextFileFunc;

type
  TMobileMenuDic = TDictionary<TListBoxItem, TMenuItem>;


  TForm1 = class(TForm)
    RichViewEdit1: TRichViewEdit;
    StatusBar1: TStatusBar;
    MainMenu1: TMainMenu;
    mpdInsert: TMenuItem;
    mitPicture: TMenuItem;
    mpdComponent: TMenuItem;
    mitButtonComp: TMenuItem;
    mitEditBoxComp: TMenuItem;
    mitBreak: TMenuItem;
    OpenDialog1: TOpenDialog;
    mpdFile: TMenuItem;
    N1: TMenuItem;
    mitExit: TMenuItem;
    mpdBullet: TMenuItem;
    mitHelpIcon: TMenuItem;
    mitHelpSearchIcon: TMenuItem;
    mitPropertiesIcon: TMenuItem;
    mitSave: TMenuItem;
    SaveDialog1: TSaveDialog;
    N2: TMenuItem;
    mitClear: TMenuItem;
    mpdEdit: TMenuItem;
    mitCopy: TMenuItem;
    mitPaste: TMenuItem;
    mitCut: TMenuItem;
    mitDelete: TMenuItem;
    PasteAs1: TMenuItem;
    mitPasteAsText: TMenuItem;
    mitPasteAsBitmap: TMenuItem;
    mitPasteAsRVF: TMenuItem;
    N3: TMenuItem;
    mitEditCheckpoint: TMenuItem;
    mitEditProps: TMenuItem;
    PopupMenu1: TPopupMenu;
    mitEditProp1: TMenuItem;
    mitEditCheckpoint1: TMenuItem;
    mpdHotspot: TMenuItem;
    mitAddImageHS: TMenuItem;
    mitAddTextHS: TMenuItem;
    mitSelectAll: TMenuItem;
    mpdMisc: TMenuItem;
    N4: TMenuItem;
    mitPrint: TMenuItem;
    N6: TMenuItem;
    mitPreview: TMenuItem;
    mitSelectCurrentWord: TMenuItem;
    mpdBackground: TMenuItem;
    mitBackNoBitmap: TMenuItem;
    mitBackStretched: TMenuItem;
    mitBackTiled: TMenuItem;
    mitBackTiledandScrolled: TMenuItem;
    mitBackCentered: TMenuItem;
    N5: TMenuItem;
    mitCheckpointList: TMenuItem;
    N9: TMenuItem;
    mitRemovePageBreak: TMenuItem;
    mitInsertPageBreak: TMenuItem;
    N10: TMenuItem;
    mitUndo: TMenuItem;
    mitRedo: TMenuItem;
    mitInsertFile: TMenuItem;
    mitLoad: TMenuItem;
    mpdTable: TMenuItem;
    mitInserttable1: TMenuItem;
    mitInsertTable2: TMenuItem;
    N7: TMenuItem;
    mitMergeCells: TMenuItem;
    N13: TMenuItem;
    mitUmRows: TMenuItem;
    mitUmCols: TMenuItem;
    mitUmRowsandCols: TMenuItem;
    Insert1: TMenuItem;
    mitRowsAbove: TMenuItem;
    mitRowsBelow: TMenuItem;
    N14: TMenuItem;
    mitColsLeft: TMenuItem;
    mitColsRight: TMenuItem;
    Delete1: TMenuItem;
    mitDelRows: TMenuItem;
    mitDelColumns: TMenuItem;
    Unmerge1: TMenuItem;
    Split1: TMenuItem;
    mitSplitVertically: TMenuItem;
    mitSplitHorizontally: TMenuItem;
    mitInsertTable3: TMenuItem;
    mitInsertTable4: TMenuItem;
    mitPasteAsRTF: TMenuItem;
    RVStyle1: TRVStyle;
    N11: TMenuItem;
    mitReadOnly: TMenuItem;
    N12: TMenuItem;
    mitOptions: TMenuItem;
    mpdLists: TMenuItem;
    mitApplyList: TMenuItem;
    mitRemoveLists: TMenuItem;
    il: TImageList;
    mitBackTopLeft: TMenuItem;
    mitBackTopRight: TMenuItem;
    mitBackBottomLeft: TMenuItem;
    mitBackBottomRight: TMenuItem;
    lblStatus: TLabel;
    OpenDialog2: TOpenDialog;
    Layout1: TLayout;
    cmbPara: TComboBox;
    cmbText: TComboBox;
    btnMenu: TSpeedButton;
    Popup1: TPopup;
    TabControl1: TTabControl;
    tabMain: TTabItem;
    lstMenuMain: TListBox;
    liGroupMainMenu: TListBoxGroupHeader;
    liFile: TListBoxItem;
    liEdit: TListBoxItem;
    liInsert: TListBoxItem;
    liList: TListBoxItem;
    liTable: TListBoxItem;
    liMisc: TListBoxItem;
    tabEdit: TTabItem;
    lstMenuEdit: TListBox;
    liGroupEdit: TListBoxGroupHeader;
    liEditUndo: TListBoxItem;
    liEditRedo: TListBoxItem;
    liGroupEditSel: TListBoxGroupHeader;
    liEditCut: TListBoxItem;
    liEditCopy: TListBoxItem;
    liEditPaste: TListBoxItem;
    liEditDelete: TListBoxItem;
    liEditSelectAll: TListBoxItem;
    liEditSelectWord: TListBoxItem;
    liGroupEditPB: TListBoxGroupHeader;
    liEditPageBreakIns: TListBoxItem;
    liEditPageBreakDel: TListBoxItem;
    liGroupEditObjects: TListBoxGroupHeader;
    liEditCheckpoint: TListBoxItem;
    liEditProperties: TListBoxItem;
    tabInsert: TTabItem;
    lstMenuInsert: TListBox;
    liGroupFromFile: TListBoxGroupHeader;
    liInsDoc: TListBoxItem;
    liInsImage: TListBoxItem;
    liGroupInsComponent: TListBoxGroupHeader;
    liInsButton: TListBoxItem;
    liInsEdit: TListBoxItem;
    liGroupInsBullet: TListBoxGroupHeader;
    liInsBullet2: TListBoxItem;
    liInsBullet3: TListBoxItem;
    liGroupInsHotspot: TListBoxGroupHeader;
    liInsHotspot2: TListBoxItem;
    liGroupInsTable: TListBoxGroupHeader;
    liInsTable1: TListBoxItem;
    liInsTable2: TListBoxItem;
    liInsTable3: TListBoxItem;
    liInsTable4: TListBoxItem;
    liGroupInsOther: TListBoxGroupHeader;
    liInsBreak: TListBoxItem;
    tabList: TTabItem;
    lstMenuList: TListBox;
    ListBoxGroupHeader12: TListBoxGroupHeader;
    liListApply: TListBoxItem;
    liListRemove: TListBoxItem;
    tabTable: TTabItem;
    lstMenuTable: TListBox;
    liGroupTableInsert: TListBoxGroupHeader;
    liTableRowAbove: TListBoxItem;
    liTableRowBelow: TListBoxItem;
    liTableColLeft: TListBoxItem;
    liTableColRight: TListBoxItem;
    liGroupTableDelete: TListBoxGroupHeader;
    liTableDelRows: TListBoxItem;
    liTableDelCols: TListBoxItem;
    liGroupTableMerge: TListBoxGroupHeader;
    liTableMerge: TListBoxItem;
    liGroupTableSplit: TListBoxGroupHeader;
    liTableSplitVert: TListBoxItem;
    liTableSplitHorz: TListBoxItem;
    liGroupTableUnmerge: TListBoxGroupHeader;
    liTableUnmergeRows: TListBoxItem;
    liTableUnmergeCols: TListBoxItem;
    liTableUnmergeAll: TListBoxItem;
    tabMisc: TTabItem;
    lstMenuMisc: TListBox;
    liGroupMiscBackgroundDemo: TListBoxGroupHeader;
    liMiscColor: TListBoxItem;
    liMiscStretched: TListBoxItem;
    liMiscTiled: TListBoxItem;
    liMiscTiledAndScrolled: TListBoxItem;
    liMiscCenter: TListBoxItem;
    liMiscTopLeft: TListBoxItem;
    liMiscTopRight: TListBoxItem;
    liMiscBottomLeft: TListBoxItem;
    liMiscBottomRight: TListBoxItem;
    liGroupMiscOther: TListBoxGroupHeader;
    liMiscGoToCheckpoint: TListBoxItem;
    liMiscReadOnly: TListBoxItem;
    tabFile: TTabItem;
    lstMenuFile: TListBox;
    ListBoxGroupHeader1: TListBoxGroupHeader;
    liFileLoad: TListBoxItem;
    liFileSave: TListBoxItem;
    liFileOptions: TListBoxItem;
    liFileClear: TListBoxItem;
    GridPanelLayout1: TGridPanelLayout;
    liInsBullet1: TListBoxItem;
    liInsHotSpot1: TListBoxItem;
    GlowEffect1: TGlowEffect;
    procedure FormCreate(Sender: TObject);
    procedure mitPictureClick(Sender: TObject);
    procedure mitButtonCompClick(Sender: TObject);
    procedure mitEditBoxCompClick(Sender: TObject);
    procedure RichViewEdit1CurParaStyleChanged(Sender: TObject);
    procedure RichViewEdit1CurTextStyleChanged(Sender: TObject);
    procedure mitBreakClick(Sender: TObject);
    procedure mitExitClick(Sender: TObject);
    procedure mitInsertBulletClick(Sender: TObject);
    procedure mitSaveClick(Sender: TObject);
    procedure mitClearClick(Sender: TObject);
    procedure mpdEditClick(Sender: TObject);
    procedure mitPasteAsBitmapClick(Sender: TObject);
    procedure mitPasteAsTextClick(Sender: TObject);
    procedure mitPasteClick(Sender: TObject);
    procedure mitDeleteClick(Sender: TObject);
    procedure mitCutClick(Sender: TObject);
    procedure mitCopyClick(Sender: TObject);
    procedure mitEditCheckpointClick(Sender: TObject);
    procedure mitAddHSClick(Sender: TObject);
    procedure mitSelectAllClick(Sender: TObject);
    procedure mitEditPropsClick(Sender: TObject);
    procedure RichViewEdit1RVFControlNeeded(Sender: TCustomRichView;
      Name: TRVUnicodeString; const Tag: TRVTag; var ctrl: TControl);
    procedure RichViewEdit1RVFImageListNeeded(Sender: TCustomRichView;
      ImageListTag: Integer; var il: TCustomImageList);
    procedure mitCheckPointListClick(Sender: TObject);
    procedure mitPasteAsRVFClick(Sender: TObject);
    procedure PopupMenu1Popup(Sender: TObject);
    procedure mitBackClick(Sender: TObject);
    procedure RichViewEdit1SaveComponentToFile(Sender: TCustomRichView;
      Path: TRVUnicodeString; SaveMe: TPersistent; SaveFormat: TRVSaveFormat;
      var OutStr: TRVUnicodeString);
    procedure mitSelectCurrentWordClick(Sender: TObject);
    procedure RichViewEdit1Jump(Sender: TObject; id: Integer);
    procedure mpdBackgroundClick(Sender: TObject);
    procedure mitInsertPageBreakClick(Sender: TObject);
    procedure mitRemovePageBreakClick(Sender: TObject);
    procedure mitUndoClick(Sender: TObject);
    procedure mitRedoClick(Sender: TObject);
    procedure mitInsertFileClick(Sender: TObject);
    procedure mitPasteAsUnicodeTextClick(Sender: TObject);
    procedure mitLoadClick(Sender: TObject);
    procedure mitInserttable1Click(Sender: TObject);
    procedure mitInsertTable2Click(Sender: TObject);
    procedure mitCellsOperationClick(Sender: TObject);
    procedure mpdTableClick(Sender: TObject);
    procedure mitInsertTable3Click(Sender: TObject);
    procedure mitInsertTable4Click(Sender: TObject);
    procedure RichViewEdit1RVMouseMove(Sender: TObject; id: Integer);
    procedure mitPasteAsRTFClick(Sender: TObject);
    procedure mitReadOnlyClick(Sender: TObject);
    procedure mitOptionsClick(Sender: TObject);
    procedure mitApplyListClick(Sender: TObject);
    procedure mitRemoveListsClick(Sender: TObject);
    procedure RichViewEdit1ControlAction(Sender: TCustomRichView;
      ControlAction: TRVControlAction; ItemNo: Integer; var ctrl: TControl);
    procedure RichViewEdit1RVFPictureNeeded(Sender: TCustomRichView;
      const ItemName: TRVUnicodeString; Item: TRVNonTextItemInfo; Index1,
      Index2: Integer; var gr: TRVGraphicFM);
    procedure cmbTextChange(Sender: TObject);
    procedure cmbParaChange(Sender: TObject);
    procedure liMainMenuClick(Sender: TObject);
    procedure btnMenuClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; var KeyChar: Char;
      Shift: TShiftState);
    procedure lstMenuItemClick(const Sender: TCustomListBox;
      const Item: TListBoxItem);
    procedure FormDestroy(Sender: TObject);
    procedure RichViewEdit1ImportPicture(Sender: TCustomRichView;
      const Location: TRVUnicodeString; Width, Height: Integer;
      var Graphic: TRVGraphicFM);
    procedure RichViewEdit1Click(Sender: TObject);
    procedure RVStyle1DrawStyleText(Sender: TRVStyle; const s: TRVUnicodeString;
      Canvas: TRVFMXCanvas; StyleNo: Integer; SpaceAtLeft, ALeft, ATop, AWidth,
      AHeight, DrawOffsY, TextWidth: Single; SpaceCount: Integer;
      DrawState: TRVTextDrawStates; Printing, PreviewCorrection,
      ForMetafile: Boolean; ColorMode: TRVColorMode; DarkMode: Boolean;
      DefBiDiMode: TRVBiDiMode; RefCanvas: TRVFMXCanvas; LeftNoExpIndex,
      RightNoExpIndex: Integer; PSaD: PRVScreenAndDevice;
      var DoDefault: Boolean);
    procedure RichViewEdit1WriteHyperlink(Sender: TCustomRichView; id: Integer;
      RVData: TCustomRVData; ItemNo: Integer; SaveFormat: TRVSaveFormat;
      var Target, Extras: TRVUnicodeString);
    procedure RichViewEdit1SaveItemToFile(Sender: TCustomRichView;
      const Path: TRVUnicodeString; RVData: TCustomRVData; ItemNo: Integer;
      SaveFormat: TRVSaveFormat; var OutStr: TRVUnicodeString;
      var DoDefault: Boolean);
  private
    { Private declarations }
    FFillCombos: Boolean;
    MobileMenuDic: TMobileMenuDic;
    procedure OnControlClick(Sender: TObject);
    procedure UpdateUndoMenu(out UndoText, RedoText: String;
      out UndoEnabled, RedoEnabled: Boolean);
    function GetRVFErrors: String;
    procedure FillStyleCombo(Styles: TCustomRVInfos; cmb: TComboBox);
    procedure BuildMobileMenuDic;
    procedure RVMClosePopup(var Msg: TRVMessage);
    {$IFDEF ANDROID}
    procedure SaveFile_ToURL(const URL, MimeType: TRVUnicodeString);
    {$ENDIF}
    procedure ClearDocument;
    procedure SaveFile_ToFileName(const FileName: TRVUnicodeString);
    procedure OpenFile_FromStream(Stream: TStream; const MimeType: TRVUnicodeString);
    procedure OpenFile_FromFileName(const FileName: TRVUnicodeString);
    procedure InsertFile_FromStream(Stream: TStream; const MimeType: TRVUnicodeString);
    procedure InsertFile_FromFileName(const FileName: TRVUnicodeString);
    procedure InsertGraphic(Graphic: TRVGraphicFM);
    {$IFDEF RVMOBILE}
    procedure DoOpenFile(Sender: TObject; RequestCode: Integer;
      const URL, DisplayName, MimeType, FileName: TRVUnicodeString; Content: TStream);
    procedure DoCreateFile(Sender: TObject; RequestCode: Integer;
      const URL, DisplayName, MimeType, FileName: TRVUnicodeString);
    procedure DoOpenImage(Sender: TObject; Bmp: TRVBitmap);
    procedure DoFileError(Sender: TObject; RequestCode: Integer);
    {$ENDIF}
    procedure DoAfterLoadFile(Success: Boolean);
    procedure UpdateStyleComboSelection;
    procedure AdjustShortcuts;
//    function LoadResourceFontByName( const ResourceName : string; ResType: PChar ) : Boolean;
  public
  end;

var
  Form1: TForm1;

implementation

uses
  CPFrm, PropFrm, ListFrm, OptionsFrm, RVUndoStr, System.StrUtils,
  FMX.FontManager, FMX.Skia;

const
  IMAGE_SIZE = 23;
  RVM_CLOSEPOPUP = 1;
  CODE_OPENDOCUMENT = 100;
  CODE_OPENIMAGE    = 101;
  CODE_SAVEDOCUMENT = 102;
  CODE_INSERTDOCUMENT = 103;
  ALL_FORMATS_FILTER = '*.rvf;*.rtf;*.docx;*.htm;*.html;*.markdown;*.md;*.txt';

{$R *.FMX}

function GetDataFolder: String;
begin
  {$IFDEF MSWINDOWS}
  Result := ExtractFilePath(ParamStr(0))+'..\..\';
  {$ENDIF}
  {$IFDEF OSX}
  // files are in "Data" folder in the folder of the app bundle
  Result := ExtractFilePath(NSStrToStr(TNSBundle.Wrap(TNSBundle.OCClass.mainBundle).bundlePath)) +
    'Data/';
  {$ENDIF}
  {$IFDEF ANDROID}
  Result := System.IOUtils.TPath.GetDocumentsPath + PathDelim;
  {$ENDIF}
  {$IFDEF IOS}
  Result := System.IOUtils.TPath.GetDocumentsPath + PathDelim;
  {$ENDIF}
  {$IFDEF LINUX}
  Result := ExtractFilePath(ParamStr(0));
  {$ENDIF}
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  Randomize;
  AdjustShortcuts;

  {$IFDEF RVMOBILE}
  RVFileChooser.Control := RichViewEdit1;
  RVFileChooser.OnOpenFile := DoOpenFile;
  RVFileChooser.OnCreateFile := DoCreateFile;
  RVFileChooser.OnOpenImage := DoOpenImage;
  RVFileChooser.OnError := DoFileError;
  {$ELSE}
  btnMenu.Visible := False;
  {$ENDIF}
  Popup1.Visible := False;
  RVStyle1.ListStyles[0].Levels[0].Font.Charset := SYMBOL_CHARSET;

  // Loading the initial file.
  RichViewEdit1.AddNL('Meal options:', 0, 0);
  RichViewEdit1.LoadRVF(GetDataFolder + 'Readme.rvf');
  FillStyleCombo(RVStyle1.ParaStyles, cmbPara);
  FillStyleCombo(RVStyle1.TextStyles, cmbText);
  RichViewEdit1.Format;
  UpdateStyleComboSelection;

  MobileMenuDic := TMobileMenuDic.Create;
  BuildMobileMenuDic;
  RVMessagingSystem.RegisterMessageHandler(Self, RVM_CLOSEPOPUP, RVMClosePopup);

  var vFontFile := GetDataFolder + 'Wingdings.ttf';
  if FileExists(vFontFile) then
    TSkDefaultProviders.RegisterTypeface(vFontFile)
  else
    ShowMessage('Font File not found: ' + vFontFile);

//  var LFontManager: IFMXFontManagerService;
//  if TPlatformServices.Current.SupportsPlatformService(IFMXFontManagerService, LFontManager) then
//  begin
//    if FileExists(vFontFile) then
//      LFontManager.AddCustomFontFromFile(vFontFile)
//    else
//      ShowMessage('Font File not found: ' + vFontFile);
//  end;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  {$IFDEF RVMOBILE}
  RVFileChooser.OnOpenFile := nil;
  RVFileChooser.OnCreateFile := nil;
  RVFileChooser.OnOpenImage := nil;
  RVFileChooser.OnError := nil;
  {$ENDIF}
  RVMessagingSystem.UnRegister(Self);
  FreeAndNil(MobileMenuDic);
end;

procedure TForm1.FormResize(Sender: TObject);
begin
  {$IFDEF RVMOBILE}
  if Popup1.Width > Screen.Width then
    Popup1.Width := Screen.Width;
  Popup1.Height := Screen.Height * 0.8;
  {$ENDIF}
end;

procedure TForm1.UpdateStyleComboSelection;
begin
  RichViewEdit1CurTextStyleChanged(nil);
  RichViewEdit1CurParaStyleChanged(nil);
end;

procedure TForm1.AdjustShortcuts;
{$IFDEF MACOS}
var
  i: Integer;
  MI: TMenuItem;
{$ENDIF}
begin
  {$IFDEF MACOS}
  // Changing Ctrl to Cmd in all shortcuts
  for i := 0 to ComponentCount - 1 do
    if Components[i] is TMenuItem then
    begin
      MI := TMenuItem(Components[i]);
      if (MI.ShortCut and scCtrl) <> 0 then
        MI.ShortCut := (MI.ShortCut and not scCtrl) or scCommand;
    end;
  {$ENDIF}
end;


{ Filling combobox with standard styles ------------------------------}
procedure TForm1.FillStyleCombo(Styles: TCustomRVInfos; cmb: TComboBox);
var
  i: Integer;
begin
  { The simplest way to fill the combo box with style names is:
    cmb.Items.Assign(Styles);
   But this code will fill the combo box with all styles -
   both standard styles (i.e. real styles) and non-standard styles will be
   added in it.
   So we'll fill in the combo box manually.
   For simplification, we'll add only the first standard styles }
  FFillCombos := True;
  cmb.Items.BeginUpdate;
  cmb.Items.Clear;
  for i := 0 to Styles.Count - 1 do
  begin
    if not TCustomRVInfo(Styles.Items[i]).Standard then
      break;
    cmb.Items.Add(TCustomRVInfo(Styles.Items[i]).StyleName);
  end;
  cmb.Items.EndUpdate;
  FFillCombos := False;
end;

{---------------------------------------------------------------------}
procedure TForm1.UpdateUndoMenu(out UndoText, RedoText: String;
      out UndoEnabled, RedoEnabled: Boolean);
var
  UndoType: TRVUndoType;
begin
  UndoType := RichViewEdit1.UndoAction;
  UndoEnabled := UndoType <> rvutNone;
  if UndoType = rvutCustom then
    UndoText := 'Undo ' + RichViewEdit1.UndoName
  else
    UndoText := 'Undo ' + RVUndoTypeNamesEn[UndoType];

  UndoType := RichViewEdit1.RedoAction;
  RedoEnabled := UndoType <> rvutNone;
  if UndoType = rvutCustom then
    RedoText := 'Redo ' + RichViewEdit1.RedoName
  else
    RedoText := 'Redo ' + RVUndoTypeNamesEn[UndoType];
end;

{======================================================================}
{ Font and paragraph combos                                            }
{======================================================================}
procedure TForm1.RichViewEdit1CurParaStyleChanged(Sender: TObject);
begin
  FFillCombos := True;
  if RichViewEdit1.CurParaStyleNo < cmbPara.Items.Count then
    cmbPara.ItemIndex := RichViewEdit1.CurParaStyleNo
  else
    cmbPara.ItemIndex := -1;
  FFillCombos := False;
end;

{----------------------------------------------------------------------}
procedure TForm1.RichViewEdit1CurTextStyleChanged(Sender: TObject);
begin
  FFillCombos := True;
  if RichViewEdit1.CurTextStyleNo < cmbText.Items.Count then
    cmbText.ItemIndex := RichViewEdit1.CurTextStyleNo
  else
    cmbText.ItemIndex := -1;
  FFillCombos := False;
end;
procedure TForm1.RichViewEdit1ImportPicture(Sender: TCustomRichView;
  const Location: TRVUnicodeString; Width, Height: Integer;
  var Graphic: TRVGraphicFM);
begin
  ShowMessage(Format('Location=%s', [Location]));
end;

{----------------------------------------------------------------------}
procedure TForm1.cmbParaChange(Sender: TObject);
begin
  if FFillCombos then
    exit;
  RichViewEdit1.ApplyParaStyle(cmbPara.ItemIndex);
  RichViewEdit1.SetFocus;
end;

{----------------------------------------------------------------------}
procedure TForm1.cmbTextChange(Sender: TObject);
begin
  if FFillCombos then
    exit;
  RichViewEdit1.ApplyTextStyle(cmbText.ItemIndex);
  RichViewEdit1.SetFocus;
end;

{======================================================================}
{ Main menu: "File"                                                    }
{======================================================================}
function GetFileOpenFormat(const FileName: String): TRVLoadFormat;
var
  Ext: String;
begin
  Ext := ExtractFileExt(FileName).ToLower;
  if Ext = '.rvf' then
    Result := rvlfRVF
  else if Ext = '.rtf' then
    Result := rvlfRTF
  else if Ext = '.docx' then
    Result := rvlfDocX
  else if Pos('htm', Ext) <> 0 then
    Result := rvlfHTML
  else if (Ext = '.markdown') or (Ext = '.mdown') or (Ext = '.mkdn') or
    (Ext = '.md') or (Ext = '.mkd') or (Ext = '.mdwn') or (Ext = '.mdtxt') then
    Result := rvlfMarkdown
  else
    Result := rvlfText;
end;

function GetFileOpenFormatByMime(const S: String): TRVLoadFormat;
begin
  if (S = 'application/octet-stream') or (S = 'application/rvf') then
    Result := rvlfRVF
  else if (S = 'application/rtf') or (S = 'text/rtf') then
    Result := rvlfRTF
  else if S = 'application/vnd.openxmlformats-officedocument.wordprocessingml.document' then
    Result := rvlfDocX
  else if (S = 'text/html') or (S = 'application/xhtml+xml') then
    Result := rvlfHTML
  else if (S = 'text/markdown') then
    Result := rvlfMarkdown
  else
    Result := rvlfText;
end;

function GetFileSaveFormat(FileName: String): TRVSaveFormat;
var
  Ext: String;
begin
  Ext := ExtractFileExt(FileName).ToLower;
  if Ext = '.rvf' then
    Result := rvsfRVF
  else if Ext = '.rtf' then
    Result := rvsfRTF
  else if Ext = '.docx' then
    Result := rvsfDocX
  else if Pos('htm', Ext) <> 0 then
    Result := rvsfHTML
  else if (Ext = '.markdown') or (Ext = '.mdown') or (Ext = '.mkdn') or
    (Ext = '.md') or (Ext = '.mkd') or (Ext = '.mdwn') or (Ext = '.mdtxt') then
    Result := rvsfMarkdown
  else
    Result := rvsfText;
end;

function GetFileSaveFormatByMime(const S: String): TRVSaveFormat;
begin
  if (S = 'application/octet-stream') or (S = 'application/rvf') then
    Result := rvsfRVF
  else if (S = 'application/rtf') or (S = 'text/rtf') then
    Result := rvsfRTF
  else if S = 'application/vnd.openxmlformats-officedocument.wordprocessingml.document' then
    Result := rvsfDocX
  else if (S = 'text/html') or (S = 'application/xhtml+xml') then
    Result := rvsfHTML
  else if (S = 'text/markdown') then
    Result := rvsfMarkdown
  else
    Result := rvsfText;
end;
{ File|Load... --------------------------------------------------------}

procedure TForm1.mitLoadClick(Sender: TObject);
begin
  lblStatus.Text := '';
  {$IFDEF RVMOBILE}
  // Android, iOS
  RVFileChooser.OpenFile('*/*', ALL_FORMATS_FILTER, 'Open Document', CODE_OPENDOCUMENT);
  {$ELSE}
  // Windows, macOS, Linux
  OpenDialog1.Title := 'Load File';
  if OpenDialog1.Execute then
    OpenFile_FromFileName(OpenDialog1.FileName);
  {$ENDIF}
end;
{ File|Save... --------------------------------------------------------}
procedure TForm1.mitSaveClick(Sender: TObject);
//
//  procedure ProcessHyperlinks();
//  begin
//    for var i := RichViewEdit1.RVData.Items.Count-1 downto 0 do
//    begin
//      var vItemInfo := RichViewEdit1.RVData.GetItem(i);
//      var vTarget := vItemInfo.Tag;
//      var sPos := Pos('(', vTarget);
//      if sPos > 0 then
//      begin
//        var ePos := Pos(')', vTarget, sPos);     // a(12)
//        if ePos > sPos then
//        begin
//          var vStyleNo := StrToIntDef(Copy(vTarget, sPos+1, ePos - sPos - 1), -1);
//          if (vStyleNo >= 0) and (vStyleNo <> vItemInfo.StyleNo) then
//          begin
//            vItemInfo.StyleNo := vStyleNo;
//          end;
//        end;
//      end;
//    end;
//  end;
//
begin
  {$IFDEF RVMOBILE}
  // Android, iOS
  RVFileChooser.CreateFile('*/*', ALL_FORMATS_FILTER, 'Save As', 'File.rvf', CODE_SAVEDOCUMENT);
  {$ELSE}
  // Windows, macOS, Linux
  if SaveDialog1.Execute then
  begin
//    ProcessHyperlinks();
    SaveFile_ToFileName(SaveDialog1.FileName);
  end;
  {$ENDIF}
end;

{ Event: called by RVFileChooser after RVFileChooser.OpenFile or GetFileContent }
procedure TForm1.DoAfterLoadFile(Success: Boolean);
var
  ErrorMessage: String;
begin
  if not Success then
  begin
    ErrorMessage := 'Error during loading';
    if OpenDialog1.FilterIndex = 1 then
      ErrorMessage := ErrorMessage + GetRVFErrors;
    RVShowError(ErrorMessage);
  end;
  FillStyleCombo(RVStyle1.ParaStyles, cmbPara);
  FillStyleCombo(RVStyle1.TextStyles, cmbText);
  RichViewEdit1.DeleteUnusedStyles(true, true, true);
  RichViewEdit1.Format;
  UpdateStyleComboSelection;
end;
{------------------------------------------------------------------------------}
procedure TForm1.ClearDocument;
begin
  RichViewEdit1.LeftMargin := 5;
  RichViewEdit1.RightMargin := 5;
  RichViewEdit1.TopMargin := 5;
  RichViewEdit1.BottomMargin := 5;
  RichViewEdit1.Clear;
end;
{------------------------------------------------------------------------------}
{$IFDEF ANDROID}
procedure TForm1.SaveFile_ToURL(const URL, MimeType: TRVUnicodeString);
var
  Content: TMemoryStream;
  r: Boolean;
begin
  Content := TMemoryStream.Create;
  try
    case GetFileSaveFormatByMime(MimeType) of
      rvsfRVF:
        r := RichViewEdit1.SaveRVFToStream(Content, False);
      rvsfDocX:
        r := RichViewEdit1.SaveDocXToStream(Content, False);
      rvsfRTF:
        r := RichViewEdit1.SaveRTFToStream(Content, False);
      rvsfHTML:
        r := RichViewEdit1.SaveHTMLToStream(Content);
      rvsfMarkdown:
        r := RichViewEdit1.SaveMarkdownToStream(Content, '', False);
      rvsfText:
        r := SaveEncodedTextToStream(RichViewEdit1, '', Content,
          TEncoding.UTF8);
      else
        r := False;
    end;
    if r then
    begin
      Content.Position := 0;
      r := RVFileChooser.WriteFile(URL, Content);
    end;
  finally
    Content.Free;
  end;
  if not r then
    RVShowError('Error during saving');
end;
{$ENDIF}
{------------------------------------------------------------------------------}
procedure TForm1.SaveFile_ToFileName(const FileName: TRVUnicodeString);
var
  r: Boolean;
begin
  case GetFileSaveFormat(FileName) of
    rvsfRVF:
      r := RichViewEdit1.SaveRVF(FileName, False);
    rvsfDocX:
      r := RichViewEdit1.SaveDocX(FileName, False);
    rvsfRTF:
      r := RichViewEdit1.SaveRTF(FileName, False);
    rvsfHTML:
      r := RichViewEdit1.SaveHTML(FileName);
    rvsfMarkdown:
      r := RichViewEdit1.SaveMarkdown(FileName, False);
    rvsfText:
      r := SaveEncodedTextToFile(RichViewEdit1, FileName,
        TEncoding.UTF8);
    else
      r := False;
  end;
  if not r then
    RVShowError('Error during saving');
end;
{------------------------------------------------------------------------------}
procedure TForm1.OpenFile_FromStream(Stream: TStream; const MimeType: TRVUnicodeString);
var
  r: Boolean;
  Enc: TEncoding;
begin
  ClearDocument;
  case GetFileOpenFormatByMime(MimeType) of
    rvlfRVF:
      r := RichViewEdit1.LoadRVFFromStream(Stream);
    rvlfDocX:
      r := RichViewEdit1.LoadDocXFromStream(Stream);
    rvlfRTF:
      r := RichViewEdit1.LoadRTFFromStream(Stream);
    rvlfHTML:
      r := RichViewEdit1.LoadHTMLFromStream(Stream);
    rvlfMarkdown:
      r := RichViewEdit1.LoadMarkdownFromStream(Stream);
    rvlfText:
      r := LoadTextFromStreamAuto(RichViewEdit1, Stream, Enc);
    else
      r := False;
  end;
  DoAfterLoadFile(r);
end;
{------------------------------------------------------------------------------}
procedure TForm1.OpenFile_FromFileName(const FileName: TRVUnicodeString);
var
  r: Boolean;
  Enc: TEncoding;
begin
  ClearDocument;
  case GetFileOpenFormat(FileName) of
    rvlfRVF:
      r := RichViewEdit1.LoadRVF(FileName);
    rvlfDocX:
      r := RichViewEdit1.LoadDocX(FileName);
    rvlfRTF:
      r := RichViewEdit1.LoadRTF(FileName);
    rvlfHTML:
      r := RichViewEdit1.LoadHTML(FileName);
    rvlfMarkdown:
      r := RichViewEdit1.LoadMarkdown(FileName);
    rvlfText:
      r := LoadTextFromFileAuto(RichViewEdit1, FileName, Enc);
    else
      r := False;
  end;
  DoAfterLoadFile(r);
end;
{------------------------------------------------------------------------------}
procedure TForm1.InsertFile_FromStream(Stream: TStream; const MimeType: TRVUnicodeString);
var
  r: Boolean;
begin
  case GetFileOpenFormatByMime(MimeType) of
    rvlfRVF:
      r := RichViewEdit1.InsertRVFFromStreamEd(Stream);
    rvlfDocX:
      r := RichViewEdit1.InsertDocXFromStreamEd(Stream);
    rvlfRTF:
      r := RichViewEdit1.InsertRTFFromStreamEd(Stream);
    rvlfHTML:
      r := RichViewEdit1.InsertHTMLFromStreamEd(Stream);
    rvlfMarkdown:
      r := RichViewEdit1.InsertMarkdownFromStreamEd(Stream);
    rvlfText:
      r := InsertTextFromStreamAuto(RichViewEdit1, Stream);
    else
      r := False;
  end;
  if not r then
    RVShowError('Error reading file');
end;
{------------------------------------------------------------------------------}
procedure TForm1.InsertFile_FromFileName(const FileName: TRVUnicodeString);
var
  r: Boolean;
begin
  case GetFileOpenFormat(FileName) of
    rvlfRVF:
      r := RichViewEdit1.InsertRVFFromFileEd(FileName);
    rvlfDocX:
      r := RichViewEdit1.InsertDocXFromFileEd(FileName);
    rvlfRTF:
      r := RichViewEdit1.InsertRTFFromFileEd(FileName);
    rvlfHTML:
      r := RichViewEdit1.InsertHTMLFromFileEd(FileName);
    rvlfMarkdown:
      r := RichViewEdit1.InsertMarkdownFromFileEd(FileName);
    rvlfText:
      r := InsertTextFromFileAuto(RichViewEdit1, FileName);
    else
      r := False;
  end;
  if not r then
    RVShowError('Error reading file');
end;
{------------------------------------------------------------------------------}
procedure TForm1.InsertGraphic(Graphic: TRVGraphicFM);
begin
  if (Graphic <> nil) and ((Graphic.Width = 0) or (Graphic.Height = 0)) then
    FreeAndNil(Graphic);
  if Graphic <> nil then
  begin
    if Graphic.Width <= RichViewEdit1.Width then
       RichViewEdit1.InsertPicture('', Graphic, rvvaAbsMiddle)
    else
      RichViewEdit1.InsertPicture('', Graphic, rvvaAbsMiddle,
        Round(RichViewEdit1.Width),
        Round(Graphic.Height * RichViewEdit1.Width / Graphic.Width));
  end
  else
    RVShowError('Cannot open image file')
end;
{------------------------------------------------------------------------------}
{$IFDEF RVMOBILE}
{ Event: called by RVFileChooser after RVFileChooser.OpenFile, GetFileContent,
  GetImageFileContent }
procedure TForm1.DoOpenFile(Sender: TObject; RequestCode: Integer;
  const URL, DisplayName, MimeType, FileName: TRVUnicodeString; Content: TStream);
begin
  RichViewEdit1.SetFocus;
  if Content <> nil then
  begin
    // Android
    Content.Position := 0;
    case RequestCode of
      CODE_OPENDOCUMENT:
        begin
          ClearDocument;
          RichViewEdit1.Format;
          RichViewEdit1.Repaint;
          OpenFile_FromStream(Content, MimeType);
        end;
      CODE_INSERTDOCUMENT:
        InsertFile_FromStream(Content, MimeType);
      CODE_OPENIMAGE:
        // normally, an image is inserted in DoOpenImage
        // this is a fallback code, if the application does not have
        // rights to access external storage (and use media library service)
        InsertGraphic(RVGraphicHandler.LoadFromStream(Content));
    end;
  end
  else
  begin
    // iOS
    case RequestCode of
      CODE_OPENDOCUMENT:
        OpenFile_FromFileName(FileName);
      CODE_INSERTDOCUMENT:
        InsertFile_FromFileName(FileName);
      CODE_OPENIMAGE:
        // normally, an image is inserted in DoOpenImage
        // this is a fallback code, if the application does not have
        // rights to access external storage (and use media library service)
        InsertGraphic(RVGraphicHandler.LoadFromFile(FileName));
    end;
  end
end;
{ Event: called by RVFileChooser after RVFileChooser.CreateFile }
procedure TForm1.DoCreateFile(Sender: TObject; RequestCode: Integer;
  const URL, DisplayName, MimeType, FileName: TRVUnicodeString);
begin
  case RequestCode of
    CODE_SAVEDOCUMENT:
      begin
        {$IFDEF ANDROID}
        // Android
        SaveFile_ToURL(URL, MimeType);
        {$ELSE}
        // iOS
        SaveFile_ToFileName(FileName);
        {$ENDIF}
      end;
  end;
end;
{ Event: called by RVFileChooser on opening image }
procedure TForm1.DoOpenImage(Sender: TObject; Bmp: TRVBitmap);
begin
  InsertGraphic(Bmp);
end;

{ Event: called by RVFileChooser on error }
procedure TForm1.DoFileError(Sender: TObject; RequestCode: Integer);
begin
  case RequestCode of
    CODE_OPENDOCUMENT, CODE_OPENIMAGE, CODE_INSERTDOCUMENT:
      RVShowError('Cannot open file');
    CODE_SAVEDOCUMENT:
      RVShowError('Cannot save file');
  end;
end;

{$ENDIF}

{ Event: picture needed while reading from RVF ------------------------}

// the event itself
procedure TForm1.RichViewEdit1RVFPictureNeeded(Sender: TCustomRichView;
  const ItemName: TRVUnicodeString; Item: TRVNonTextItemInfo; Index1,
  Index2: Integer; var gr: TRVGraphicFM);
var
  FileName: String;
begin
  if Item is TRVGraphicItemInfo then
  begin
    FileName := 'default.bmp';
    if FileName = '' then
      exit;
    gr := TRVBitmap.Create;
    gr.LoadFromFile(GetDataFolder + FileName);
  end;
end;

{ Event: control needed while reading from RVF ------------------------}
procedure TForm1.RichViewEdit1WriteHyperlink(Sender: TCustomRichView;
  id: Integer; RVData: TCustomRVData; ItemNo: Integer;
  SaveFormat: TRVSaveFormat; var Target, Extras: TRVUnicodeString);
begin
  Target := RVData.GetItemTag(ItemNo);
  // Strip off any StyleNo reference
  var sPos := Pos('(', Target);
  if sPos > 0 then
  begin
    var ePos := Pos(')', Target, sPos);
    if ePos > sPos then
    begin
      Delete(Target, sPos, ePos - sPos + 1);
    end;
  end;
end;

procedure TForm1.RichViewEdit1RVFControlNeeded(Sender: TCustomRichView;
  Name: TRVUnicodeString; const Tag: TRVTag; var ctrl: TControl);
begin
  ctrl := TButton.Create(RichViewEdit1);
  TButton(ctrl).Text := 'from file';
end;

{ Event: imagelist needed while reading from RVF ----------------------}
procedure TForm1.RichViewEdit1RVFImageListNeeded(Sender: TCustomRichView;
  ImageListTag: Integer; var il: TCustomImageList);
begin
  il := Self.il;
end;

{---------------------------------------------------------------------}
function TForm1.GetRVFErrors: String;
begin
  Result := '';
  if rvfwUnknownPicFmt in RichViewEdit1.RVFWarnings then
    Result := Result + 'unknown picture format;';
  if rvfwUnknownCtrls in RichViewEdit1.RVFWarnings then
    Result := Result + 'unknown control class;';
  if rvfwConvUnknownStyles in RichViewEdit1.RVFWarnings then
    Result := Result + 'text, paragraph or list style is not present;';
  if rvfwConvLargeImageIdx in RichViewEdit1.RVFWarnings then
    Result := Result + 'invalid image-list index;';
  if rvfwUnknownStyleProperties in RichViewEdit1.RVFWarnings then
    Result := Result + 'unknown property of text, paragraph or list style;';
  if Result <> '' then
    Result := #13'(' + Result + ')';
end;

{ File|Options... --------------------------------------------------------}

procedure TForm1.mitOptionsClick(Sender: TObject);
begin
  frmOptions.SetOptions(RichViewEdit1);
  {$IFDEF RVMOBILE}
  frmOptions.Show;
  {$ELSE}
  frmOptions.ShowModal;
  {$ENDIF}
end;

{ Event: saving controls in HTML --------------------------------------}
// Note: not all browsers support <INPUT> tags outside <FORM></FORM> tags
procedure TForm1.RichViewEdit1SaveComponentToFile(Sender: TCustomRichView;
  Path: TRVUnicodeString; SaveMe: TPersistent; SaveFormat: TRVSaveFormat;
  var OutStr: TRVUnicodeString);
begin
  case SaveFormat of
    rvsfText:
      begin
        OutStr := '(' + SaveMe.ClassName + ')';
      end;
    rvsfHTML:
      begin
        if SaveMe is TButton then
        begin
          OutStr := '<input type="button" value="' + TButton(SaveMe).Text +
            '" ' + 'onClick="alert(''Just a demo'')">';
          exit;
        end;
        if SaveMe is TEdit then
        begin
          OutStr := '<input type="text" value="' + TEdit(SaveMe).Text + '">';
          exit;
        end;
      end;
    rvsfRTF:
      begin
        OutStr := '{\plain\b (' + SaveMe.ClassName + ')}';
      end;
  end;
end;

procedure TForm1.RichViewEdit1SaveItemToFile(Sender: TCustomRichView;
  const Path: TRVUnicodeString; RVData: TCustomRVData; ItemNo: Integer;
  SaveFormat: TRVSaveFormat; var OutStr: TRVUnicodeString;
  var DoDefault: Boolean);
begin

end;

{ File|Clear ----------------------------------------------------------}
procedure TForm1.mitClearClick(Sender: TObject);
begin
  ClearDocument;
  RichViewEdit1.Format;
  UpdateStyleComboSelection;
end;

{ File|Exit -----------------------------------------------------------}
procedure TForm1.mitExitClick(Sender: TObject);
begin
  Close;
end;

{======================================================================}
{ Main menu: "Insert"                                                  }
{======================================================================}
{ Insert|File... ------------------------------------------------------}
procedure TForm1.mitInsertFileClick(Sender: TObject);
begin
  {$IFDEF RVMOBILE}
  RVFileChooser.GetFileContent('*/*', ALL_FORMATS_FILTER, 'Insert File', CODE_INSERTDOCUMENT);
  {$ELSE}
  OpenDialog1.Title := 'Insert File';
  if OpenDialog1.Execute then
    InsertFile_FromFileName(OpenDialog1.FileName);
  {$ENDIF}
end;

{ Insert|Picture... ---------------------------------------------------}
procedure TForm1.mitPictureClick(Sender: TObject);
begin
  {$IFDEF RVMOBILE}
  RVFileChooser.GetImageFileContent('Insert Image', CODE_OPENIMAGE);
  {$ELSE}
  OpenDialog2.Title := 'Inserting Image';
  OpenDialog2.Filter := RVGraphicHandler.GetFileDialogFilter;
  if OpenDialog2.Execute then
    InsertGraphic(RVGraphicHandler.LoadFromFile(OpenDialog2.FileName));
  {$ENDIF}
end;

{ Event: clicking inserted control ------------------------------------}
procedure TForm1.OnControlClick(Sender: TObject);
begin
  RichViewEdit1.SelectControl(TControl(Sender));
end;

{ Insert|Component|Button ---------------------------------------------}
procedure TForm1.mitButtonCompClick(Sender: TObject);
var
  btn: TButton;
const
  Captions: array [0 .. 9] of String = ('Help', 'Exit', 'Cancel', 'Ok', 'Close',
    'Run', 'Options...', 'Minimize', 'Hide', 'Show');
begin
  btn := TButton.Create(Self);
  btn.Text := Captions[Random(10)];
  btn.OnClick := OnControlClick;
  if RichViewEdit1.InsertControl('', btn, rvvaBaseLine) then
    RichViewEdit1.SetCurrentItemExtraIntProperty(rvepResizable, 1, True);
end;

{ Insert|Component|Edit Box -------------------------------------------}
procedure TForm1.mitEditBoxCompClick(Sender: TObject);
var
  edt: TEdit;
const
  Captions: array [0 .. 9] of String = ('0', 'Hello', '1', '$0', '2x2=4',
    'enter text here', 'x<y', 'to be or not to be?', '(empty)', '(full)');
begin
  edt := TEdit.Create(Self);
  edt.Text := Captions[Random(10)];
  edt.OnClick := OnControlClick;
  RichViewEdit1.InsertControl('', edt, rvvaBaseLine);
  if RichViewEdit1.CurItemStyle = rvsComponent then
    RichViewEdit1.SetCurrentItemExtraIntProperty(rvepResizable, 1, True);
end;

{ Insert|Bullet|"XXX" -------------------------------------------------}
procedure TForm1.mitInsertBulletClick(Sender: TObject);
begin
  RichViewEdit1.InsertBullet(TControl(Sender).Tag, il, IMAGE_SIZE, IMAGE_SIZE);
end;

{ Insert|Hot Spot|"XXX" -----------------------------------------------}
procedure TForm1.mitAddHSClick(Sender: TObject);
begin
  RichViewEdit1.InsertHotSpot(TControl(Sender).Tag,
    TMenuItem(Sender).Tag + 2, il, IMAGE_SIZE, IMAGE_SIZE);
end;

{ Insert|Break --------------------------------------------------------}
procedure TForm1.mitBreakClick(Sender: TObject);
begin
  RichViewEdit1.InsertBreak(1, rvbsLine, TAlphaColorRec.Null);
end;

{======================================================================}
{ Main menu : "Edit"                                                   }
{======================================================================}
{ Edit ----------------------------------------------------------------}
procedure TForm1.mpdEditClick(Sender: TObject);
var
  UndoText, RedoText: String;
  UndoEnabled, RedoEnabled: Boolean;
begin
  mitPasteAsRTF.Enabled := RichViewEdit1.CanPasteRTF;
  mitPasteAsText.Enabled := RVClipboardHasTextW;
  mitPasteAsBitmap.Enabled := RVClipboardHasBitmap;
  mitPasteAsRVF.Enabled := RichViewEdit1.CanPasteRVF;
  mitPaste.Enabled := RichViewEdit1.CanPaste;

  mitCopy.Enabled := RichViewEdit1.SelectionExists;
  mitCut.Enabled := mitCopy.Enabled;

  mitInsertPageBreak.Enabled := (RichViewEdit1.InplaceEditor = nil);
  mitRemovePageBreak.Enabled := (RichViewEdit1.InplaceEditor = nil) and
    RichViewEdit1.PageBreaksBeforeItems[RichViewEdit1.CurItemNo];

  // You can edit properties only for item at the caret position.
  // We disable this item because otherwise user can think what he will
  // edit properties of all selected items.
  // More smart programs can determine if there is only one item is selected
  // and do not disable this item in this case
  mitEditProps.Enabled := not RichViewEdit1.SelectionExists;

  UpdateUndoMenu(UndoText, RedoText, UndoEnabled, RedoEnabled);
  mitUndo.Enabled := UndoEnabled;
  mitRedo.Enabled := RedoEnabled;
  mitUndo.Text := UndoText;
  mitRedo.Text := RedoText;
end;

{ Edit|Undo------------------------------------------------------------}
procedure TForm1.mitUndoClick(Sender: TObject);
begin
  RichViewEdit1.Undo;
end;

{ Edit|Redo -----------------------------------------------------------}
procedure TForm1.mitRedoClick(Sender: TObject);
begin
  RichViewEdit1.Redo;
end;

{ Edit|Cut ------------------------------------------------------------}
procedure TForm1.mitCutClick(Sender: TObject);
begin
  RichViewEdit1.CutDef;
end;

{ Edit|Copy -----------------------------------------------------------}
procedure TForm1.mitCopyClick(Sender: TObject);
begin
  RichViewEdit1.CopyDef;
end;

{ Edit|Paste ----------------------------------------------------------}
procedure TForm1.mitPasteClick(Sender: TObject);
begin
  RichViewEdit1.Paste;
end;

{ Edit|Paste As|RTF ---------------------------------------------------}
procedure TForm1.mitPasteAsRTFClick(Sender: TObject);
begin
  RichViewEdit1.PasteRTF;
end;

{ Edit|Paste As|Text --------------------------------------------------}
procedure TForm1.mitPasteAsTextClick(Sender: TObject);
begin
  RichViewEdit1.PasteText;
end;

{ Edit|Paste As|Unicode Text ------------------------------------------}
procedure TForm1.mitPasteAsUnicodeTextClick(Sender: TObject);
begin
  RichViewEdit1.PasteTextW;
end;

{ Edit|Paste As|Bitmap ------------------------------------------------}
procedure TForm1.mitPasteAsBitmapClick(Sender: TObject);
begin
  RichViewEdit1.PasteBitmap(False);
end;

{ Edit|Paste As|RichView Format ---------------------------------------}
procedure TForm1.mitPasteAsRVFClick(Sender: TObject);
begin
  RichViewEdit1.PasteRVF;
end;
{-----------------------------------------------------------------------}
procedure TForm1.RichViewEdit1Click(Sender: TObject);
begin
  var vTag := String(RichViewEdit1.GetCurrentTag);

  if vTag.StartsWith('$noprint:{0}', true) then
    RichViewEdit1.SetCurrentTag(ReplaceStr(vTag, '{0}', '{1}'))

  else if vTag.StartsWith('$noprint:{1}', true) then
    RichViewEdit1.SetCurrentTag(ReplaceStr(vTag, '{1}', '{0}'));
end;

procedure TForm1.RichViewEdit1ControlAction(Sender: TCustomRichView;
  ControlAction: TRVControlAction; ItemNo: Integer; var ctrl: TControl);
begin
  if ControlAction = rvcaAfterRVFLoad then
  begin
    if ctrl is TButton then
      TButton(ctrl).OnClick := OnControlClick
    else if ctrl is TEdit then
      TEdit(ctrl).OnClick := OnControlClick
  end;
end;

{ Edit|Delete ---------------------------------------------------------}
procedure TForm1.mitDeleteClick(Sender: TObject);
begin
  // Shortcut to this item is Ctrl+Del
  // If you make it Del, you will be unable to use del key in editor
  RichViewEdit1.DeleteSelection;
end;

{ Edit|Select All -----------------------------------------------------}
procedure TForm1.mitSelectAllClick(Sender: TObject);
begin
  { warning: SelectAll moves caret to the end of the text }
  RichViewEdit1.SelectAll;
  RichViewEdit1.SetFocus;
  RichViewEdit1.Repaint;
end;

{ Edit| Insert Page Break----------------------------------------------}
procedure TForm1.mitInsertPageBreakClick(Sender: TObject);
begin
  RichViewEdit1.InsertPageBreak;
end;

{ Edit| Remove Page Break----------------------------------------------}
procedure TForm1.mitRemovePageBreakClick(Sender: TObject);
begin
  RichViewEdit1.RemoveCurrentPageBreak;
end;

{ Edit|Checkpoint... --------------------------------------------------}
procedure TForm1.mitEditCheckpointClick(Sender: TObject);
var
  CpNo:           Integer;
  Tag:            TRVTag;
  Name:           TRVUnicodeString;
  CheckPointData: TCheckPointData;
  RaiseEvent:     Boolean;
begin
  CheckPointData := RichViewEdit1.GetCurrentCheckpoint;
  if CheckPointData <> nil then
  begin
    RichViewEdit1.GetCheckpointInfo(CheckPointData, Tag, Name, RaiseEvent);
    CpNo := RichViewEdit1.GetCheckpointNo(CheckPointData);
    frmCp.lblStatus.Text := 'Editing checkpoint #' + IntToStr(CpNo);
    frmCp.txtName.Text := Name;
    frmCp.txtTag.Text := Tag;
    frmCp.btnOk.Text := 'OK';
    frmCp.btnBack.StyleLookup := 'backtoolbutton';
    frmCp.btnDelete.Enabled := True;
    frmCp.btnDelete1.Visible := True;
  end
  else
  begin
    frmCp.lblStatus.Text := 'New checkpoint';
    frmCp.txtName.Text := '';
    frmCp.txtTag.Text := '';
    frmCp.btnOk.Text := 'Add';
    frmCp.btnBack.StyleLookup := 'addtoolbutton';
    frmCp.btnDelete.Enabled := False;
    frmCp.btnDelete1.Visible := False;
  end;
  {$IFDEF RVMOBILE}
  frmCp.Show;
  {$ELSE}
  frmCp.ShowModal;
  {$ENDIF}
end;

{ Edit|Select Current Word -------------------------------------}
procedure TForm1.mitSelectCurrentWordClick(Sender: TObject);
begin
  RichViewEdit1.SelectCurrentWord;
  // now you can do something with current word:
  // translate or spell check, for example...
end;

{ Edit|Current Item Properties... -------------------------------------}
procedure TForm1.mitEditPropsClick(Sender: TObject);
begin
  frmProp.GetFromEditor(RichViewEdit1);
  {$IFDEF RVMOBILE}
  frmProp.Show;
  {$ELSE}
  frmProp.ShowModal;
  {$ENDIF}
end;

{======================================================================}
{ Main menu : "Misc"                                                   }
{======================================================================}
{ Misc | Go to checkpoint ... -----------------------------------------}
procedure TForm1.mitCheckPointListClick(Sender: TObject);
var
  X, Y:           Single;
  Tag:            TRVTag;
  Name:           TRVUnicodeString;
  CheckPointData: TCheckPointData;
  RaiseEvent:     Boolean;
  s:              String;
begin
  { Does not work for checkpoints in table cells }
  frmList.lst.Items.Clear;
  CheckPointData := RichViewEdit1.GetFirstCheckPoint;
  while CheckPointData <> nil do
  begin
    RichViewEdit1.GetCheckpointInfo(CheckPointData, Tag, Name, RaiseEvent);
    RichViewEdit1.GetCheckpointXY(CheckPointData, X, Y);
    s := Format('(X:%f,Y:%f) Name:"%s" Tag:"%s"', [X, Y, Name, Tag]);
    frmList.lst.Items.Add(s);
    CheckPointData := RichViewEdit1.GetNextCheckpoint(CheckPointData);
  end;
  {$IFDEF RVMOBILE}
  frmList.Show;
  {$ELSE}
  frmList.ShowModal;
  {$ENDIF}
end;

{ Misc | Read-Only -----------------------------------------------------}
procedure TForm1.mitReadOnlyClick(Sender: TObject);
begin
  RichViewEdit1.ReadOnly := not RichViewEdit1.ReadOnly;
  mitReadOnly.IsChecked := RichViewEdit1.ReadOnly;
  if RichViewEdit1.ReadOnly then
    liMiscReadOnly.ItemData.Accessory := TListBoxItemData.TAccessory.aCheckmark
  else
    liMiscReadOnly.ItemData.Accessory := TListBoxItemData.TAccessory.aNone;

end;

{ Misc | Background submenu popups ------------------------------------}
procedure TForm1.mpdBackgroundClick(Sender: TObject);
begin
  // Displaying RichViewEdit1.BackgroundStyle as checkmark in submenu...
  mitBackNoBitmap.IsChecked := RichViewEdit1.BackgroundStyle = bsNoBitmap;
  mitBackStretched.IsChecked := RichViewEdit1.BackgroundStyle = bsStretched;
  mitBackTiledandScrolled.IsChecked :=
    RichViewEdit1.BackgroundStyle = bsTiledAndScrolled;
  mitBackTiled.IsChecked := RichViewEdit1.BackgroundStyle = bsTiled;
  mitBackCentered.IsChecked := RichViewEdit1.BackgroundStyle = bsCentered;
  mitBackTopLeft.IsChecked := RichViewEdit1.BackgroundStyle = bsTopLeft;
  mitBackTopRight.IsChecked := RichViewEdit1.BackgroundStyle = bsTopRight;
  mitBackBottomLeft.IsChecked := RichViewEdit1.BackgroundStyle = bsBottomLeft;
  mitBackBottomRight.IsChecked := RichViewEdit1.BackgroundStyle = bsBottomRight;
end;

{ Misc | Background options -------------------------------------------}
procedure TForm1.mitBackClick(Sender: TObject);
begin
  RichViewEdit1.BackgroundStyle := TBackgroundStyle(TMenuItem(Sender).Tag);
end;

{======================================================================}
{ On Popup -------------------------------------------------------------}
procedure TForm1.PopupMenu1Popup(Sender: TObject);
begin
  mitEditProp1.Enabled := not RichViewEdit1.SelectionExists;
end;

{-----------------------------------------------------------------------}
{ Event: OnJump (when user clicks hypertext item with pressed Ctrl key   }
procedure TForm1.RichViewEdit1Jump(Sender: TObject; id: Integer);
var
  RVData: TCustomRVFormattedData;
  ItemNo: Integer;
  s:      String;
begin
  // NOTE: OnJump is called after the caret is repositioned to clicked item
  // But warning: a clicked item is not necessarily an active item!
  // (when clicking on left part of a picture or left part of the first character
  // a text item,
  // the caret moves before the item and the previous item becomes active!)
  RichViewEdit1.GetJumpPointLocation(id, RVData, ItemNo);
  s := RVData.GetItemTag(ItemNo);
  //s := RVData.GetItemText(ItemNo);
  RVShowMessage(Format('Tag of clicked hyperlink is "%s"', [s]));
end;

{------------------------------------------------------------------------------}
{ Event: OnRVMouseMove (when user moves mouse above hypertext item with pressed Ctrl key   }
procedure TForm1.RichViewEdit1RVMouseMove(Sender: TObject; id: Integer);
var
  RVData: TCustomRVFormattedData;
  ItemNo: Integer;
  s:      String;
begin
  if id = -1 then
  begin
    lblStatus.Text := '';
  end
  else
  begin
    RichViewEdit1.GetJumpPointLocation(id, RVData, ItemNo);
    s := RVData.GetItemTag(ItemNo);
    lblStatus.Text := Format('Tag of hyperlink is "%s"', [s]);
  end;
end;

{======================================================================}
{ Main menu : "Lists"                                                  }
{======================================================================}
{ Lists | Apply -------------------------------------------------------}
procedure TForm1.mitApplyListClick(Sender: TObject);
begin
  // See more demos about list styles in Demos\Delphi\Assorted\ListStyles\
  if (RVStyle1.ListStyles.Count = 0) or (RVStyle1.ListStyles[0].Levels.Count = 0)
  then
  begin
    RVShowError('Default list style is not defined');
    exit;
  end;
  RichViewEdit1.ApplyListStyle(0, 0, 1, False, False);
end;

{ Lists | Remove ------------------------------------------------------}
procedure TForm1.mitRemoveListsClick(Sender: TObject);
begin
  RichViewEdit1.RemoveLists(False);
end;

{======================================================================}
{ Main menu : "Table"                                                  }
{======================================================================}
{ Table | Insert Table Example 1 --------------------------------------}
procedure TForm1.mitInserttable1Click(Sender: TObject);
var
  table: TRVTableItemInfo;
  r, c:  Integer;
begin
  table := TRVTableItemInfo.CreateEx(4, 3, RichViewEdit1.RVData);

  table.BorderStyle := rvtbRaisedColor;
  table.CellBorderStyle := rvtbLoweredColor;
  table.BorderLightColor := $FFC9F1FA;
  table.BorderColor := $FF108EA9;
  table.CellBorderLightColor := $FFC9F1FA;
  table.CellBorderColor := $FF108EA9;
  table.Color := $FF24C7EA;
  table.BorderWidth := 5;
  table.CellBorderWidth := 2;
  table.CellPadding := 5;
  table.CellVSpacing := 1;
  table.CellHSpacing := 1;
  table.BorderVSpacing := 1;
  table.BorderHSpacing := 1;

  for r := 0 to table.RowCount - 1 do
    for c := 0 to table.ColCount - 1 do
      table.Cells[r, c].BestWidth := 100;

  table.MergeCells(0, 0, 3, 1, False);
  table.MergeCells(1, 0, 1, 3, False);
  with table.Cells[0, 0] do
  begin
    Color := TAlphaColorRec.Bisque;
    Clear;
    AddBullet('', 0, il, 2);
    AddNL(' Example 1 ', 1, -1);
    AddBullet('', 0, il, -1);
    AddNL('All cells have 100 pixels width, width of table itself is calculated basing on width of cells.',
      0, 0);
  end;

  if RichViewEdit1.InsertItem('', table) then
  begin
  end;
end;

{ Table | Insert Table Example 2 --------------------------------------}
procedure TForm1.mitInsertTable2Click(Sender: TObject);
var
  table: TRVTableItemInfo;
  btn:   TButton;
begin
  table := TRVTableItemInfo.CreateEx(10, 6, RichViewEdit1.RVData);
  table.Color := TAlphaColorRec.White;

  table.BorderStyle := rvtbRaisedColor;
  table.CellBorderStyle := rvtbLoweredColor;
  table.BorderLightColor := TAlphaColorRec.White;
  table.BorderColor := TAlphaColorRec.Black;
  table.CellBorderLightColor := TAlphaColorRec.White;
  table.CellBorderColor := TAlphaColorRec.Black;

  table.BorderWidth := 2;
  table.BorderVSpacing := 0;
  table.BorderHSpacing := 0;
  table.CellBorderWidth := 2;
  table.CellPadding := 3;
  table.CellVSpacing := 0;
  table.CellHSpacing := 0;
  table.Cells[0, 0].BestWidth := -16;
  table.Cells[0, 1].BestWidth := -16;
  table.Cells[0, 2].BestWidth := -16;
  table.Cells[0, 3].BestWidth := -16;
  table.Cells[0, 4].BestWidth := -16;
  table.Cells[0, 5].BestWidth := -16;
  // table.Rows.MergeCells(1,0,6,1);
  table.MergeCells(2, 0, 2, 8, False);
  with table.Cells[2, 0] do
  begin
    Clear;
    AddNL('Another example.', 0, 0);
    btn := TButton.Create(nil);
    btn.Text := 'With button inside';
    btn.Width := 150;
    btn.OnClick := OnControlClick;
    AddControl('', btn, 2, rvvaBaseLine);
    SetItemExtraIntProperty(ItemCount - 1, rvepResizable, 1);
    AddNL('Width of table = 90% of document width. Widths of cells = 16%', 0,
      0);
  end;
  table.BestWidth := -90;
  if RichViewEdit1.InsertItem('', table) then
  begin
  end;
end;

{ Table | Insert Table Example 3 --------------------------------------}
procedure TForm1.mitInsertTable3Click(Sender: TObject);
var
  table: TRVTableItemInfo;
  r, c:  Integer;
begin
  table := TRVTableItemInfo.CreateEx(5, 6, RichViewEdit1.RVData);

  table.Color := $FFE7CCA5;
  table.BorderStyle := rvtbColor;
  table.CellBorderStyle := rvtbColor;
  table.BorderColor := $FF34122E;
  table.CellBorderColor := $FF34122E;

  table.BorderWidth := 2;
  table.BorderVSpacing := 2;
  table.BorderHSpacing := 2;
  table.CellBorderWidth := 1;
  table.CellPadding := 3;
  table.CellVSpacing := 0;
  table.CellHSpacing := 0;
  table.Options := table.Options + [rvtoHideGridLines];

  for c := 0 to table.ColCount - 1 do
    table.Cells[0, c].Color := $FFF8E1A5;

  for r := 1 to table.RowCount - 1 do
    table.Cells[r, 0].Color := $FFF8E1A5;

  for r := 1 to table.RowCount - 1 do
    for c := 1 to table.ColCount - 1 do
    begin
      table.Cells[r, c].Color := $FFDAB47A;
      if c > 1 then
        table.Cells[r, c].VisibleBorders.Left := False;
      if c < table.Rows[r].Count - 1 then
        table.Cells[r, c].VisibleBorders.Right := False;
    end;

  table.BestWidth := 400;
  RichViewEdit1.InsertText
    ('Third example: width of table = 400 pixels, widths of cells - unspecified.',
    False);
  if RichViewEdit1.InsertItem('', table) then
  begin
  end;
end;

{ Table | Insert Table Example 4 --------------------------------------}
procedure TForm1.mitInsertTable4Click(Sender: TObject);
var
  table: TRVTableItemInfo;
  r, c:  Integer;
begin
  table := TRVTableItemInfo.CreateEx(3, 3, RichViewEdit1.RVData);

  table.Color := TAlphaColorRec.Null;
  table.BorderStyle := rvtbColor;
  table.CellBorderStyle := rvtbColor;

  table.BorderWidth := 1;
  table.BorderVSpacing := 2;
  table.BorderHSpacing := 2;
  table.CellBorderWidth := 1;
  table.CellPadding := 3;
  table.CellVSpacing := 5;
  table.CellHSpacing := 5;
  table.VRuleWidth := 1;
  table.HRuleWidth := 1;
  for r := 0 to table.RowCount - 1 do
    for c := 0 to table.ColCount - 1 do
    begin
      table.Cells[r, c].BestWidth := 40;
      table.Cells[r, c].Clear;
      table.Cells[r, c].AddFmt('%d,%d', [r, c], 0, 0);
      table.Cells[r, c].Color := TAlphaColorRec.White;
    end;
  RichViewEdit1.InsertText('Transparent table with rules', False);
  if RichViewEdit1.InsertItem('', table) then
  begin
  end;
end;

{ Table submenu popups ------------------------------------------}
procedure TForm1.mpdTableClick(Sender: TObject);
var
  Item:                           TCustomRVItemInfo;
  table:                          TRVTableItemInfo;
  r, c, cs, rs:                   Integer;
  rve:                            TCustomRichViewEdit;
  Selected, SelectionRectangular: Boolean;
begin
  if not RichViewEdit1.GetCurrentItemEx(TRVTableItemInfo, rve, Item) then
  begin
    mitRowsAbove.Enabled := False;
    mitRowsBelow.Enabled := False;
    mitColsLeft.Enabled := False;
    mitColsRight.Enabled := False;
    mitDelRows.Enabled := False;
    mitDelColumns.Enabled := False;
    mitMergeCells.Enabled := False;
    mitUmRows.Enabled := False;
    mitUmCols.Enabled := False;
    mitUmRowsandCols.Enabled := False;
    mitSplitVertically.Enabled := False;
    mitSplitHorizontally.Enabled := False;
    exit;
  end;
  table := TRVTableItemInfo(Item);
  Selected := table.GetNormalizedSelectionBounds(True, r, c, cs, rs);
  mitRowsAbove.Enabled := Selected;
  mitRowsBelow.Enabled := Selected;
  mitColsLeft.Enabled := Selected;
  mitColsRight.Enabled := Selected;
  mitDelRows.Enabled := Selected;
  mitDelColumns.Enabled := Selected;
  mitMergeCells.Enabled := table.CanMergeSelectedCells(True);
  SelectionRectangular := Selected and (table.CanMergeSelectedCells(True) or
    (table.GetEditedCell(r, c) <> nil));
  mitSplitVertically.Enabled := SelectionRectangular;
  mitSplitHorizontally.Enabled := SelectionRectangular;
  mitUmRows.Enabled := SelectionRectangular;
  mitUmCols.Enabled := SelectionRectangular;
  mitUmRowsandCols.Enabled := SelectionRectangular;
end;

{ Table | All other commands --------------------------------------}
procedure TForm1.mitCellsOperationClick(Sender: TObject);
var
  Item:         TCustomRVItemInfo;
  table:        TRVTableItemInfo;
  Data:         Single;
  r, c, cs, rs: Integer;
  s:            String;
  rve:          TCustomRichViewEdit;
  ItemNo:       Integer;
begin
  if not RichViewEdit1.CanChange or not RichViewEdit1.GetCurrentItemEx
    (TRVTableItemInfo, rve, Item) then
    exit;
  table := TRVTableItemInfo(Item);
  ItemNo := rve.GetItemNo(table);
  rve.BeginItemModify(ItemNo, Data);
  case TControl(Sender).Tag of
    1:
      table.InsertRowsAbove(1);
    2:
      table.InsertRowsBelow(1);
    3:
      table.InsertColsLeft(1);
    4:
      table.InsertColsRight(1);
    5:
      begin
        table.GetNormalizedSelectionBounds(True, r, c, cs, rs);
        if rs = table.Rows.Count then
        begin
          // deleting the whole table
          rve.SetSelectionBounds(ItemNo, 0, ItemNo, 1);
          rve.DeleteSelection;
          exit;
        end;
        rve.BeginUndoGroup(rvutModifyItem);
        rve.SetUndoGroupMode(True);
        table.DeleteSelectedRows;
        // it's possible all-nil rows/cols appear after deleting
        table.DeleteEmptyRows;
        table.DeleteEmptyCols;
        rve.SetUndoGroupMode(False);
      end;
    6:
      begin
        table.GetNormalizedSelectionBounds(True, r, c, cs, rs);
        if cs = table.Rows[0].Count then
        begin
          // deleting the whole table
          rve.SetSelectionBounds(ItemNo, 0, ItemNo, 1);
          rve.DeleteSelection;
          exit;
        end;
        rve.BeginUndoGroup(rvutModifyItem);
        rve.SetUndoGroupMode(True);
        table.DeleteSelectedCols;
        // it's possible all-nil rows/cols appear after deleting
        table.DeleteEmptyRows;
        table.DeleteEmptyCols;
        rve.SetUndoGroupMode(False);
      end;
    7:
      begin
        // 3 methods: MergeSelectedCells, DeleteEmptyRows, DeleteEmptyCols
        // must be undone as one action.
        // So using BeginUndoGroup - SetUndoGroupMode(True) - ... - SetUndoGroupMode(False)
        rve.BeginUndoGroup(rvutModifyItem);
        rve.SetUndoGroupMode(True);
        table.MergeSelectedCells(True);
        table.DeleteEmptyRows;
        table.DeleteEmptyCols;
        rve.SetUndoGroupMode(False);
        // table.MergeSelectedCells(False) will not allow to create empty columns
        // or rows
      end;
    8:
      table.UnmergeSelectedCells(True, False);
    9:
      table.UnmergeSelectedCells(False, True);
    10:
      table.UnmergeSelectedCells(True, True);
    11:
      begin
        s := '2';
        {$IFnDEF RVMOBILE}
        if InputQuery('Split Vertically', 'Columns (in each selected cell):', s)
        then
        {$ENDIF}
        begin
          table.SplitSelectedCellsVertically(StrToIntDef(s, 0));
        end;
      end;
    12:
      begin
        s := '2';
        {$IFnDEF RVMOBILE}
        if InputQuery('Split Horizontally', 'Rows (in each selected cell):', s)
        then
        {$ENDIF}
        begin
          table.SplitSelectedCellsHorizontally(StrToIntDef(s, 0));
        end;
      end;
  end;
  rve.EndItemModify(ItemNo, Data);
  rve.Change;
end;
{=============================== Mobile Menu ==================================}
// Mapping listbox items to menu items. It will be used in lstInsertItemClick
procedure TForm1.BuildMobileMenuDic;
begin
  // file
  MobileMenuDic.Add(liFileLoad, mitLoad);
  MobileMenuDic.Add(liFileSave, mitSave);
  MobileMenuDic.Add(liFileOptions, mitOptions);
  MobileMenuDic.Add(liFileClear, mitClear);
  // edit
  MobileMenuDic.Add(liEditUndo, mitUndo);
  MobileMenuDic.Add(liEditRedo, mitRedo);
  MobileMenuDic.Add(liEditCut, mitCut);
  MobileMenuDic.Add(liEditCopy, mitCopy);
  MobileMenuDic.Add(liEditPaste, mitPaste);
  MobileMenuDic.Add(liEditDelete, mitDelete);
  MobileMenuDic.Add(liEditSelectAll, mitSelectAll);
  MobileMenuDic.Add(liEditSelectWord, mitSelectCurrentWord);
  MobileMenuDic.Add(liEditPageBreakIns, mitInsertPageBreak);
  MobileMenuDic.Add(liEditPageBreakDel, mitRemovePageBreak);
  MobileMenuDic.Add(liEditCheckpoint, mitEditCheckpoint);
  MobileMenuDic.Add(liEditProperties, mitEditProps);
  // insert
  MobileMenuDic.Add(liInsDoc, mitInsertFile);
  MobileMenuDic.Add(liInsImage, mitPicture);
  MobileMenuDic.Add(liInsButton, mitButtonComp);
  MobileMenuDic.Add(liInsEdit, mitEditBoxComp);
  MobileMenuDic.Add(liInsBullet1, mitHelpIcon);
  MobileMenuDic.Add(liInsBullet2, mitHelpSearchIcon);
  MobileMenuDic.Add(liInsBullet3, mitPropertiesIcon);
  MobileMenuDic.Add(liInsHotSpot1, mitAddImageHS);
  MobileMenuDic.Add(liInsHotSpot2, mitAddTextHS);
  MobileMenuDic.Add(liInsTable1, mitInserttable1);
  MobileMenuDic.Add(liInsTable2, mitInserttable2);
  MobileMenuDic.Add(liInsTable3, mitInserttable3);
  MobileMenuDic.Add(liInsTable4, mitInserttable4);
  MobileMenuDic.Add(liInsBreak, mitBreak);
  // list
  MobileMenuDic.Add(liListApply, mitApplyList);
  MobileMenuDic.Add(liListRemove, mitRemoveLists);
  // table
  MobileMenuDic.Add(liTableRowAbove, mitRowsAbove);
  MobileMenuDic.Add(liTableRowBelow, mitRowsBelow);
  MobileMenuDic.Add(liTableColLeft, mitColsLeft);
  MobileMenuDic.Add(liTableColRight, mitColsRight);
  MobileMenuDic.Add(liTableDelRows, mitDelRows);
  MobileMenuDic.Add(liTableDelCols, mitDelColumns);
  MobileMenuDic.Add(liTableMerge, mitMergeCells);
  MobileMenuDic.Add(liTableSplitVert, mitSplitVertically);
  MobileMenuDic.Add(liTableSplitHorz, mitSplitHorizontally);
  MobileMenuDic.Add(liTableUnmergeRows, mitUmRows);
  MobileMenuDic.Add(liTableUnmergeCols, mitUmCols);
  MobileMenuDic.Add(liTableUnmergeAll, mitUmRowsandCols);
  // misc
  MobileMenuDic.Add(liMiscColor, mitBackNoBitmap);
  MobileMenuDic.Add(liMiscStretched, mitBackStretched);
  MobileMenuDic.Add(liMiscTiled, mitBackTiled);
  MobileMenuDic.Add(liMiscTiledAndScrolled, mitBackTiledandScrolled);
  MobileMenuDic.Add(liMiscCenter, mitBackCentered);
  MobileMenuDic.Add(liMiscTopLeft, mitBackTopLeft);
  MobileMenuDic.Add(liMiscTopRight, mitBackTopRight);
  MobileMenuDic.Add(liMiscBottomLeft, mitBackBottomLeft);
  MobileMenuDic.Add(liMiscBottomRight, mitBackBottomRight);
  MobileMenuDic.Add(liMiscGoToCheckpoint, mitCheckpointList);
  MobileMenuDic.Add(liMiscReadOnly, mitReadOnly);
end;

{------------------------------------------------------------------------------}
procedure PrepareListBox(LB: TListBox);
begin
  LB.ScrollToItem(LB.ItemByIndex(0));
  LB.ItemIndex := -1;
end;
{------------------------------------------------------------------------------}
procedure TForm1.btnMenuClick(Sender: TObject);
var
  Svc: IFMXVirtualKeyboardService;
begin
  if not Popup1.IsOpen then
  begin
    PrepareListBox(lstMenuMain);
    TabControl1.ActiveTab := tabMain;
    if TPlatformServices.Current.SupportsPlatformService(
       IFMXVirtualKeyboardService, Svc) then
      Svc.HideVirtualKeyboard;
  end;
  Popup1.IsOpen := not Popup1.IsOpen
end;
{------------------------------------------------------------------------------}
procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word; var KeyChar: Char;
  Shift: TShiftState);
begin
  if ((Key = vkHardwareBack) or (Key = vkEscape)) and Popup1.IsOpen then
  begin
    if TabControl1.ActiveTab = tabMain then
      Popup1.IsOpen := False
    else
    begin
      PrepareListBox(lstMenuMain);
      TabControl1.ActiveTab := tabMain;
    end;
    Key := 0;
  end;
end;
{------------------------------------------------------------------------------}
procedure TForm1.RVMClosePopup(var Msg: TRVMessage);
begin
  Popup1.IsOpen := False;
end;

procedure TForm1.RVStyle1DrawStyleText(Sender: TRVStyle;
  const s: TRVUnicodeString; Canvas: TRVFMXCanvas; StyleNo: Integer;
  SpaceAtLeft, ALeft, ATop, AWidth, AHeight, DrawOffsY, TextWidth: Single;
  SpaceCount: Integer; DrawState: TRVTextDrawStates; Printing,
  PreviewCorrection, ForMetafile: Boolean; ColorMode: TRVColorMode;
  DarkMode: Boolean; DefBiDiMode: TRVBiDiMode; RefCanvas: TRVFMXCanvas;
  LeftNoExpIndex, RightNoExpIndex: Integer; PSaD: PRVScreenAndDevice;
  var DoDefault: Boolean);
begin
  if rvtsSelected in DrawState then
    exit; // Use default drawing for selected text

  var vItem := TCustomRVData(Sender.RVData).GetItem(Sender.ItemNo);
  if (vItem = nil) or not (vItem is TRVTextItemInfo) then exit;

  var vTag := String(vItem.Tag);

  if vTag.StartsWith('$picklist') then
  begin
    var vPicklistStyle := RVStyle1.TextStyles[13];
    Sender.GraphicInterface.SetFontColor(Canvas, vPicklistStyle.AlphaColor);
    var p1 := TPointF.Create(ALeft, ATop + AHeight);
    var p2 := TPointF.Create(ALeft + AWidth, ATop + AHeight);
    Canvas.Canvas.Stroke.Color := vPicklistStyle.UnderlineAlphaColor;
    Canvas.Canvas.DrawLine(p1, p2, 100);
  end

  else if vTag.StartsWith('$noprint:{1}') then
  begin
    var vCloakedStyle := RVStyle1.TextStyles[14];
    Sender.GraphicInterface.SetFontColor(Canvas, vCloakedStyle.AlphaColor);
    var p1 := TPointF.Create(ALeft, ATop + AHeight);
    var p2 := TPointF.Create(ALeft + AWidth, ATop + AHeight);
    Canvas.Canvas.Stroke.Color := vCloakedStyle.UnderlineAlphaColor;
    Canvas.Canvas.DrawLine(p1, p2, 100);
  end

  else if vTag.StartsWith('$noprint:{0}') then
  begin
    var vUncloakedStyle := RVStyle1.TextStyles[15];
    var p1 := TPointF.Create(ALeft, ATop + AHeight);
    var p2 := TPointF.Create(ALeft + AWidth, ATop + AHeight);
    Canvas.Canvas.Stroke.Color := vUncloakedStyle.UnderlineAlphaColor;
    Canvas.Canvas.DrawLine(p1, p2, 100);
  end

  else if vTag.StartsWith('$checkbox') then
  begin
    var vCheckboxStyle := RVStyle1.TextStyles[16];
    var p1 := TPointF.Create(ALeft, ATop + AHeight);
    var p2 := TPointF.Create(ALeft + AWidth, ATop + AHeight);
    Canvas.Canvas.Stroke.Color := vCheckboxStyle.UnderlineAlphaColor;
    Canvas.Canvas.DrawLine(p1, p2, 100);
    var vRect := TRectF.Create(Aleft, ATop, ALeft + AWidth, ATop + AHeight);
    Sender.GraphicInterface.SetBrush(Canvas, vCheckboxStyle.BackAlphaColor);
    Canvas.Canvas.Fill.Color := vCheckboxStyle.BackAlphaColor;
    Canvas.Canvas.FillRect(vRect, 100);
  end

  else if vTag.StartsWith('$radio') then
  begin
    var vRadioStyle := RVStyle1.TextStyles[17];
    var p1 := TPointF.Create(ALeft, ATop + AHeight);
    var p2 := TPointF.Create(ALeft + AWidth, ATop + AHeight);
    Canvas.Canvas.Stroke.Color := vRadioStyle.UnderlineAlphaColor;
    Canvas.Canvas.DrawLine(p1, p2, 100);
    var vRect := TRectF.Create(Aleft, ATop, ALeft + AWidth, ATop + AHeight);
    Sender.GraphicInterface.SetBrush(Canvas, vRadioStyle.BackAlphaColor);
    Canvas.Canvas.Fill.Color := vRadioStyle.BackAlphaColor;
    Canvas.Canvas.FillRect(vRect, 100);
  end;
end;

{------------------------------------------------------------------------------}
procedure TForm1.liMainMenuClick(Sender: TObject);

  procedure CheckListBoxItem(LI: TListBoxItem; Value: Boolean);
  begin
    if Value then
      LI.ItemData.Accessory := TListBoxItemData.TAccessory.aCheckmark
    else
      LI.ItemData.Accessory := TListBoxItemData.TAccessory.aNone;
  end;

  procedure UpdateEditMenu;
  var
    UndoText, RedoText: String;
    UndoEnabled, RedoEnabled: Boolean;
  begin
    liEditPaste.Enabled := RichViewEdit1.CanPaste;
    liEditCopy.Enabled := RichViewEdit1.SelectionExists;
    liEditCut.Enabled := liEditCopy.Enabled;
    liEditDelete.Enabled := liEditCopy.Enabled;

    liEditPageBreakIns.Enabled := (RichViewEdit1.InplaceEditor = nil);
    liEditPageBreakDel.Enabled := (RichViewEdit1.InplaceEditor = nil) and
      RichViewEdit1.PageBreaksBeforeItems[RichViewEdit1.CurItemNo];

    liEditProperties.Enabled := not RichViewEdit1.SelectionExists;

    UpdateUndoMenu(UndoText, RedoText, UndoEnabled, RedoEnabled);
    liEditUndo.Enabled := UndoEnabled;
    liEditRedo.Enabled := RedoEnabled;
    liEditUndo.Text := UndoText;
    liEditRedo.Text := RedoText;
  end;

  procedure UpdateTableMenu;
  var
    Item:                           TCustomRVItemInfo;
    table:                          TRVTableItemInfo;
    r, c, cs, rs:                   Integer;
    rve:                            TCustomRichViewEdit;
    Selected, SelectionRectangular: Boolean;
  begin
    if not RichViewEdit1.GetCurrentItemEx(TRVTableItemInfo, rve, Item) then
    begin
      liTableRowAbove.Enabled := False;
      liTableRowBelow.Enabled := False;
      liTableColLeft.Enabled := False;
      liTableColRight.Enabled := False;
      liTableDelRows.Enabled := False;
      liTableDelCols.Enabled := False;
      liTableMerge.Enabled := False;
      liTableUnmergeRows.Enabled := False;
      liTableUnmergeCols.Enabled := False;
      liTableUnmergeAll.Enabled := False;
      liTableSplitVert.Enabled := False;
      liTableSplitHorz.Enabled := False;
      exit;
    end;
    table := TRVTableItemInfo(Item);
    Selected := table.GetNormalizedSelectionBounds(True, r, c, cs, rs);
    liTableRowAbove.Enabled := Selected;
    liTableRowBelow.Enabled := Selected;
    liTableColLeft.Enabled := Selected;
    liTableColRight.Enabled := Selected;
    liTableDelRows.Enabled := Selected;
    liTableDelCols.Enabled := Selected;
    liTableMerge.Enabled := table.CanMergeSelectedCells(True);
    SelectionRectangular := Selected and (table.CanMergeSelectedCells(True) or
      (table.GetEditedCell(r, c) <> nil));
    liTableSplitVert.Enabled := SelectionRectangular;
    liTableSplitHorz.Enabled := SelectionRectangular;
    liTableUnmergeRows.Enabled := SelectionRectangular;
    liTableUnmergeCols.Enabled := SelectionRectangular;
    liTableUnmergeAll.Enabled := SelectionRectangular;
  end;

  procedure UpdateMiscMenu;
  begin
    CheckListBoxItem(liMiscColor, RichViewEdit1.BackgroundStyle = bsNoBitmap);
    CheckListBoxItem(liMiscStretched, RichViewEdit1.BackgroundStyle = bsStretched);
    CheckListBoxItem(liMiscTiled, RichViewEdit1.BackgroundStyle = bsTiled);
    CheckListBoxItem(liMiscTiledAndScrolled, RichViewEdit1.BackgroundStyle = bsTiledAndScrolled);
    CheckListBoxItem(liMiscCenter, RichViewEdit1.BackgroundStyle = bsCentered);
    CheckListBoxItem(liMiscTopLeft, RichViewEdit1.BackgroundStyle = bsTopLeft);
    CheckListBoxItem(liMiscTopRight, RichViewEdit1.BackgroundStyle = bsTopRight);
    CheckListBoxItem(liMiscBottomLeft, RichViewEdit1.BackgroundStyle = bsTopLeft);
    CheckListBoxItem(liMiscBottomRight, RichViewEdit1.BackgroundStyle = bsTopRight);
  end;

begin
  case (Sender as TListBoxItem).Tag of
    0:
      begin
        PrepareListBox(lstMenuFile);
        TabControl1.ActiveTab := tabFile;
      end;
    1:
      begin
        UpdateEditMenu;
        PrepareListBox(lstMenuEdit);
        TabControl1.ActiveTab := tabEdit;
      end;
    2:
      begin
        PrepareListBox(lstMenuInsert);
        TabControl1.ActiveTab := tabInsert;
      end;
    3:
      begin
        PrepareListBox(lstMenuList);
        TabControl1.ActiveTab := tabList;
      end;
    4:
      begin
        UpdateTableMenu;
        PrepareListBox(lstMenuTable);
        TabControl1.ActiveTab := tabTable;
      end;
    5:
      begin
        UpdateMiscMenu;
        PrepareListBox(lstMenuMisc);
        TabControl1.ActiveTab := tabMisc;
      end;
  end;
end;
procedure TForm1.lstMenuItemClick(const Sender: TCustomListBox;
  const Item: TListBoxItem);
var
  MenuItem: TMenuItem;
begin
  if Item is TListBoxGroupHeader then
    exit;
  if MobileMenuDic.TryGetValue(Item, MenuItem) and Assigned(MenuItem.OnClick) then
    MenuItem.OnClick(MenuItem)
  else
    RVShowError('Menu item without command');
  RVMessagingSystem.PostMessage(Self, RVM_CLOSEPOPUP, 0, 0, False);
end;

//function TForm1.LoadResourceFontByName( const ResourceName : string; ResType: PChar ) : Boolean;
//var
//  ResStream : TResourceStream;
//  FontsCount : DWORD;
//begin
//  ResStream := TResourceStream.Create(hInstance, ResourceName, ResType);
//  try
//    Result := (AddFontMemResourceEx(ResStream.Memory, ResStream.Size, nil, @FontsCount) <> 0);
//  finally
//    ResStream.Free;
//  end;
//end;

{------------------------------------------------------------------------------}
initialization

  // We need to register classes in order to load them from rvf files
  RegisterClasses([TButton, TEdit]);

end.
